import codecs
import functools
import os
import time

import pycurl


class Download:
    def __init__(self, url):
        self.url = url
        self.filename = url.split('/')[-1]

        c = pycurl.Curl()
        c.setopt(pycurl.URL, self.url)
        c.setopt(pycurl.NOBODY, 1)
        c.perform()

        self.size = c.getinfo(pycurl.CONTENT_LENGTH_DOWNLOAD)
        c.close()
        del c

        self.chunks = []

        self.m = pycurl.CurlMulti()

        self.bom_check = {}

    def get_range(self, chunk, max_chunks):
        if chunk == max_chunks - 1:
            return f'{int((self.size // max_chunks) * chunk)}-{int(self.size - 1)}'

        return f'{int((self.size // max_chunks) * chunk)}-{int((self.size // max_chunks) * (chunk + 1) - 1)}'

    def write_body(self, buf, chunk, f):
        if not self.bom_check.get(chunk, False):
            if buf[:3] == codecs.BOM_UTF8:
                buf = buf[3:]
            self.bom_check[chunk] = True

        f.write(buf)

    def write_header(self, buf, chunk):
        print(f'{chunk}: {buf}')

    def progress(self, *args, chunk):
        ...
        # print(chunk, args)

    def download(self, chunks):
        files = []

        for chunk in range(chunks):
            files.append(open(f'{self.filename}.{chunk}', 'wb'))

            c = pycurl.Curl()
            c.setopt(pycurl.URL, self.url)
            c.setopt(pycurl.POST, 0)
            c.setopt(pycurl.IPRESOLVE, pycurl.IPRESOLVE_V4)
            c.setopt(pycurl.RANGE, self.get_range(chunk, chunks))
            c.setopt(pycurl.WRITEFUNCTION, functools.partial(self.write_body, chunk=chunk, f=files[chunk]))
            c.setopt(pycurl.HEADERFUNCTION, functools.partial(self.write_header, chunk=chunk))
            c.setopt(pycurl.PROGRESSFUNCTION, functools.partial(self.progress, chunk=chunk))
            c.setopt(pycurl.NOPROGRESS, 0)
            c.setopt(pycurl.POST, 0)
            # c.setopt(pycurl.HTTPGET, 1)
            c.setopt(pycurl.FOLLOWLOCATION, 1)
            c.setopt(pycurl.MAXREDIRS, 10)
            c.setopt(pycurl.CONNECTTIMEOUT, 30)
            c.setopt(pycurl.NOSIGNAL, 1)
            c.setopt(pycurl.SSL_VERIFYPEER, 0)
            c.setopt(pycurl.LOW_SPEED_TIME, 60)
            c.setopt(pycurl.LOW_SPEED_LIMIT, 5)

            c.setopt(
                pycurl.USERAGENT,
                b"Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/60.0.3112.113 Safari/537.36",
            )
            if pycurl.version_info()[7]:
                c.setopt(pycurl.ENCODING, b"gzip, deflate")
            c.setopt(
                pycurl.HTTPHEADER,
                [
                    b"Accept: */*",
                    b"Accept-Language: en-US,en",
                    b"Accept-Charset: ISO-8859-1,utf-8;q=0.7,*;q=0.7",
                    b"Connection: keep-alive",
                    b"Keep-Alive: 300",
                    b"Expect:",
                ],
            )
            self.chunks.append(c)

            self.m.add_handle(c)

        dones = []
        fails = []

        while True:
            while True:
                ret, num_handles = self.m.perform()
                if ret != pycurl.E_CALL_MULTI_PERFORM:
                    break

            num_q, ok_list, err_list = self.m.info_read()
            for ok in ok_list:
                dones.append(ok)
            for fail in err_list:
                fails.append(fail)
            if len(dones) + len(fails) == chunks:
                break

            self.m.select(1)
            time.sleep(0.001)

        for chunk in self.chunks:
            chunk.close()

        for file in files:
            file.close()

        f = open(self.filename, 'wb')
        for chunk in range(chunks):
            f.write(open(f'{self.filename}.{chunk}', 'rb').read())
            os.remove(f'{self.filename}.{chunk}')


def download(url, chunks):
    d = Download(url)
    d.download(chunks)
    return d



import dataclasses
import datetime
import enum
import functools
import importlib.util
import json
import os
import threading

from transitions import Machine

from joatmon.event import Event
from joatmon.utility import (
    first,
    get_module_classes,
    JSONEncoder
)


class BaseTask:
    def __init__(self, api, background=False, **kwargs):  # another parameter called cache output
        self.api = api
        self.kwargs = kwargs
        self._background = background
        if self.background:
            self.thread = threading.Thread(target=self.run)
        self.stop_event = threading.Event()

        self.events = {
            'begin': Event(),
            'end': Event(),
            'error': Event()
        }

        states = ['none', 'started', 'stopped', 'running', 'exception', 'starting', 'stopping']
        self.machine = Machine(model=self, states=states, initial='none')

    @property
    def background(self):
        return self._background

    @background.setter
    def background(self, value):
        self._background = value
        if self.background:
            self.thread = threading.Thread(target=self.run)

    @staticmethod
    def help():
        return {}

    def run(self):
        raise NotImplementedError

    def running(self):
        return not self.stop_event.is_set()

    def start(self):
        self.events['begin'].fire()

        if self.background:
            self.thread.start()
        else:
            self.run()

    def stop(self):
        self.stop_event.set()

        self.events['end'].fire()


class TaskState(enum.Enum):
    running = enum.auto()
    finished = enum.auto()


# create from json and to json methods
@dataclasses.dataclass
class TaskInfo:
    name: str
    state: TaskState
    task: BaseTask


def on_begin(name, *args, **kwargs):
    settings = json.loads(open('iva/iva.json', 'r').read())
    tasks = settings.get('tasks', [])

    task_info = first(filter(lambda x: x['name'] == name, tasks))
    if task_info is None:
        return

    task_info['last_run_time'] = datetime.datetime.now()
    if task_info['on'] == 'interval':
        task_info['next_run_time'] = task_info['last_run_time'] + datetime.timedelta(seconds=task_info['interval'])

    for idx, task in enumerate(tasks):
        if task['name'] == task_info['name']:
            tasks[idx] = task_info
            break

    settings['tasks'] = tasks
    open('iva/iva.json', 'w').write(json.dumps(settings, indent=4, cls=JSONEncoder))


def on_error(name, *args, **kwargs):
    ...


def on_end(name, *args, **kwargs):
    ...


def create(api):
    name = api.input('name')
    priority = api.input('priority')
    on = api.input('on')
    script = api.input('script')
    kwargs = {}
    for k in get_class(script).params():
        kwargs[k] = api.input(k)

    # need last run time
    # need next run time
    # need last run result
    # need interval
    # if on == 'interval' need to as for interval as well
    create_args = {
        'name': name,
        'priority': priority,
        'on': on,
        'script': script,
        'status': True,
        'kwargs': kwargs
    }

    settings = json.loads(open('iva/iva.json', 'r').read())
    tasks = settings.get('tasks', [])

    tasks.append(create_args)

    settings['tasks'] = tasks
    open('iva/iva.json', 'w').write(json.dumps(settings, indent=4, cls=JSONEncoder))


def get_class(name):
    task = None

    settings = json.loads(open('iva/iva.json', 'r').read())
    for scripts in settings.get('scripts', []):
        if os.path.isabs(scripts):
            if os.path.exists(scripts) and os.path.exists(os.path.join(scripts, f'{name}.py')):
                spec = importlib.util.spec_from_file_location(scripts, os.path.join(scripts, f'{name}.py'))
                action_module = importlib.util.module_from_spec(spec)
                spec.loader.exec_module(action_module)
            else:
                continue
        else:
            try:
                _module = __import__(scripts, fromlist=[f'{name}'])
            except ModuleNotFoundError:
                continue

            action_module = getattr(_module, name, None)

        if action_module is None:
            continue

        for class_ in get_module_classes(action_module):
            if not issubclass(class_[1], BaseTask) or class_[1] is BaseTask:
                continue

            task = class_[1]

    return task


def get(api, name, kwargs, background):
    settings = json.loads(open('iva/iva.json', 'r').read())
    task_info = first(filter(lambda x: x['status'] and x['name'] == name, settings.get('tasks', [])))

    if task_info is None:
        task_info = {'script': name, 'kwargs': {}}

    script = task_info['script']

    task = get_class(script)

    if task is None:
        api.output(f'task {name} is not found')
        return None

    kwargs = {**(kwargs or {}), **task_info['kwargs'], 'parent_os_path': api.parent_os_path, 'os_path': api.os_path}

    task = task(api, **kwargs)
    task.background = background

    task.events['begin'] += functools.partial(on_begin, name=name)
    task.events['end'] += functools.partial(on_end, name=name)
    task.events['error'] += functools.partial(on_error, name=name)
    return task



from __future__ import print_function

import json

from joatmon.assistant.task import BaseTask
from joatmon.utility import JSONEncoder


class Task(BaseTask):
    def __init__(self, api, **kwargs):
        super(Task, self).__init__(api, **kwargs)

    @staticmethod
    def help():
        return {
            "name": "contact",
            "description": "a function for user to list, create, update, search delete contact",
            "parameters": {
                "type": "object",
                "properties": {
                    "mode": {
                        "type": "string",
                        "enum": ["list", "create", "update", "search", "delete"]
                    },
                    "name": {
                        "type": "string",
                        "description": "name of the contact"
                    },
                    "email": {
                        "type": "string",
                        "description": "email address of the contact"
                    },
                    "phone": {
                        "type": "string",
                        "description": "phone number of the contact"
                    }
                },
                "required": ["mode"]
            }
        }

    def run(self):
        mode = self.kwargs.get('mode', '')

        settings = json.loads(open('iva/iva.json', 'r').read())
        contacts = settings.get('contacts', [])

        if mode == 'list':
            ...
        elif mode == 'create':
            name = self.kwargs.get('name', None) or self.api.input('what is the name of the contact')
            email = self.kwargs.get('email', None) or self.api.input('what is the email')
            phone = self.kwargs.get('phone', None) or self.api.input('what is the phone number')

            contacts.append({'name': name, 'email': email, 'phone': phone})
        elif mode == 'update':
            ...
        elif mode == 'search':
            ...

        settings['contacts'] = contacts
        open('iva/iva.json', 'w').write(json.dumps(settings, indent=4, cls=JSONEncoder))

        if not self.stop_event.is_set():
            self.stop_event.set()



import pygame
from pygame.locals import *

from joatmon.event import Event

# Colors
WHITE = (255, 255, 255)
BLACK = (0, 0, 0)
GRAY = (200, 200, 200)
RED = (255, 0, 0)


class ComboBox:
    def __init__(self):
        self._x = None
        self._y = None

        self._width = None
        self._height = None

        self._text = None
        self._font = None

        self._text_color = None
        self._background_color = None
        self._hover_color = None
        self._click_color = None

        self._border = None
        self._rect = None
        self._icon = None

        self._events = {
            'hover': Event(),
            'change': Event()
        }

        self._options = []

        self.is_hovered = False
        self.is_clicked = False

        self.is_open = False
        self.selected_option = None

    def position(self, x, y):
        self._x = x
        self._y = y

        return self

    def size(self, width, height):
        self._width = width
        self._height = height

        self._rect = pygame.Rect(self._x, self._y, self._width, self._height)

        return self

    def icon(self, icon):
        _icon = pygame.image.load(icon)

        self._width = self._height
        # self._rect = None
        self._rect = pygame.Rect(self._x, self._y, self._width, self._height)

        self._icon = pygame.transform.scale(_icon, (self._width - 10, self._height - 10))

        return self

    def text(self, text, font):
        self._text = text
        self._font = font

        return self

    def color(self, t_color, b_color, h_color, c_color):
        self._text_color = t_color
        self._background_color = b_color
        self._hover_color = h_color
        self._click_color = c_color

        return self

    def border(self, border):
        self._border = border
        return self

    def on(self, event, func):
        _event = self._events.get(event, None)
        if _event:
            _event += func

        return self

    def add_option(self, option):
        self._options.append(option)
        return self

    def remove_option(self, option):
        return self

    def handle_event(self, event):
        if event.type == pygame.MOUSEMOTION:
            self.is_hovered = self._rect.collidepoint(event.pos)
            if self.is_hovered:
                _event = self._events.get('hover', None)
                if _event:
                    _event.fire()

        if event.type == MOUSEBUTTONDOWN:
            if self._rect.collidepoint(event.pos):
                self.is_open = not self.is_open
            else:
                self.is_open = False

            # if event.type == MOUSEBUTTONUP and self.is_open:
            #     print('mouse up')

            for i, option_rect in enumerate(self.option_rects):
                if self.is_open and option_rect.collidepoint(event.pos):
                    self.selected_option = self._options[i]
                    self.is_open = False

                    _event = self._events.get('change', None)
                    if _event:
                        _event.fire()

                    break

        self.update()

    def update(self):
        self.option_rects = []

        for i, option in enumerate(self._options):
            option_rect = pygame.Rect(
                self._rect.x, self._rect.y + self._rect.height * (i + 1),
                self._rect.width, self._rect.height
            )
            self.option_rects.append(option_rect)

    def draw(self, screen):
        border_color = self._border.get('color', (0, 0, 0))
        thickness = self._border.get('thickness', 1)
        radius = self._border.get('radius', 1)
        shadow = self._border.get('shadow', 3)

        if self._border:
            border_rect = self._rect.copy()
            border_rect.x -= thickness
            border_rect.y -= thickness
            border_rect.width += 2 * thickness
            border_rect.height += 2 * thickness

            if shadow:
                shadow_rect = border_rect.copy()
                shadow_rect.x += shadow
                shadow_rect.y += shadow
                pygame.draw.rect(screen, pygame.Color('gray'), shadow_rect, border_radius=radius)  # shadow color

                pygame.draw.rect(screen, pygame.Color('gray') if self.is_hovered else pygame.Color('white'), border_rect, border_radius=radius)
            pygame.draw.rect(screen, border_color, border_rect, thickness, border_radius=radius)

        font = pygame.font.SysFont(None, 24)

        # pygame.draw.rect(screen, WHITE, self._rect)
        # pygame.draw.rect(screen, BLACK, self._rect, 2)
        #
        # pygame.draw.line(
        #    screen, BLACK, (self._rect.x + self._rect.width - 20, self._rect.y),
        #    (self._rect.x + self._rect.width - 20, self._rect.y + self._rect.height), 2
        # )

        if self.is_open:
            for i, option_rect in enumerate(self.option_rects):  # if this is the selected, change the color as well
                pygame.draw.rect(screen, GRAY, option_rect)
                pygame.draw.rect(screen, BLACK, option_rect, 1)
                text_surface = font.render(self._options[i], True, BLACK)
                screen.blit(text_surface, (option_rect.x + 5, option_rect.y + 5))

        if self.selected_option:
            text_surface = font.render(self.selected_option, True, BLACK)
            screen.blit(text_surface, (self._rect.x + 5, self._rect.y + 5))



from __future__ import print_function

import os
import shutil

from joatmon.assistant.task import BaseTask


class Task(BaseTask):
    def __init__(self, api, **kwargs):
        super(Task, self).__init__(api, **kwargs)

    @staticmethod
    def help():
        return {
            "name": "move",
            "description": "a function for user to move a file from one directory to another",
            "parameters": {
                "type": "object",
                "properties": {
                    "old": {
                        "type": "string",
                        "description": "path of the file that want to be moved"
                    },
                    "new": {
                        "type": "string",
                        "description": "new path of the file"
                    }
                },
                "required": ["old", "new"]
            }
        }

    def run(self):
        old_file = self.kwargs.get('old', '') or self.api.input('what is the old file')
        new_file = self.kwargs.get('new', '') or self.api.input('what is the new file')

        parent_os_path = self.kwargs.get('parent_os_path', '')
        os_path = self.kwargs.get('os_path', '')

        shutil.move(os.path.join(parent_os_path, os_path[1:], old_file), os.path.join(parent_os_path, os_path[1:], new_file))

        if not self.stop_event.is_set():
            self.stop_event.set()



from joatmon.plugin.core import Plugin


class CachePlugin(Plugin):
    async def add(self, key, value, duration=None):
        raise NotImplementedError

    async def get(self, key):
        raise NotImplementedError

    async def update(self, key, value):
        raise NotImplementedError

    async def remove(self, key):
        raise NotImplementedError



import asyncio
import collections
import itertools
import logging
from asyncio import (
    futures,
    InvalidStateError,
    Queue
)

from ..adapters import (
    ReaderAdapter,
    WriterAdapter
)
from ..errors import (
    HZMQTTException,
    MQTTException,
    NoDataException
)
from ..mqtt import packet_class
from ..mqtt.constants import (
    QOS_0,
    QOS_1,
    QOS_2
)
from ..mqtt.packet import (
    BAD_USERNAME_PASSWORD,
    CONNACK,
    ConnackPacket,
    CONNECT,
    CONNECTION_ACCEPTED,
    ConnectPacket,
    ConnectPayload,
    ConnectVariableHeader,
    DISCONNECT,
    DisconnectPacket,
    IDENTIFIER_REJECTED,
    MQTTFixedHeader,
    NOT_AUTHORIZED,
    PINGREQ,
    PingReqPacket,
    PINGRESP,
    PingRespPacket,
    PUBACK,
    PubackPacket,
    PUBCOMP,
    PubcompPacket,
    PUBLISH,
    PublishPacket,
    PUBREC,
    PubrecPacket,
    PUBREL,
    PubrelPacket,
    RESERVED_0,
    RESERVED_15,
    SUBACK,
    SubackPacket,
    SUBSCRIBE,
    SubscribePacket,
    UNACCEPTABLE_PROTOCOL_VERSION,
    UNSUBACK,
    UnsubackPacket,
    UNSUBSCRIBE,
    UnsubscribePacket
)
from ..session import (
    INCOMING,
    IncomingApplicationMessage,
    OUTGOING,
    OutgoingApplicationMessage,
    Session
)
from ..utils import format_client_message


class ProtocolHandlerException(BaseException):
    pass


class ProtocolHandler:
    def __init__(self, session: Session = None, loop=None):
        self.logger = logging.getLogger(__name__)
        if session:
            self._init_session(session)
        else:
            self.session = None
        self.reader = None
        self.writer = None

        if loop is None:
            self._loop = asyncio.get_event_loop()
        else:
            self._loop = loop
        self._reader_task = None
        self._keepalive_task = None
        self._reader_ready = None
        self._reader_stopped = asyncio.Event(loop=self._loop)

        self._puback_waiters = dict()
        self._pubrec_waiters = dict()
        self._pubrel_waiters = dict()
        self._pubcomp_waiters = dict()

        self._write_lock = asyncio.Lock(loop=self._loop)

    def _init_session(self, session: Session):
        assert session
        log = logging.getLogger(__name__)
        self.session = session
        self.logger = logging.LoggerAdapter(log, {'client_id': self.session.client_id})
        self.keepalive_timeout = self.session.keep_alive
        if self.keepalive_timeout <= 0:
            self.keepalive_timeout = None

    def attach(self, session, reader: ReaderAdapter, writer: WriterAdapter):
        if self.session:
            raise ProtocolHandlerException("Handler is already attached to a session")
        self._init_session(session)
        self.reader = reader
        self.writer = writer

    def detach(self):
        self.session = None
        self.reader = None
        self.writer = None

    def _is_attached(self):
        if self.session:
            return True
        else:
            return False

    async def start(self):
        if not self._is_attached():
            raise ProtocolHandlerException("Handler is not attached to a stream")
        self._reader_ready = asyncio.Event(loop=self._loop)
        self._reader_task = asyncio.Task(self._reader_loop(), loop=self._loop)
        await asyncio.wait([self._reader_ready.wait()], loop=self._loop)
        if self.keepalive_timeout:
            self._keepalive_task = self._loop.call_later(self.keepalive_timeout, self.handle_write_timeout)

        print("Handler tasks started")
        await self._retry_deliveries()
        print("Handler ready")

    async def stop(self):
        self._stop_waiters()
        if self._keepalive_task:
            self._keepalive_task.cancel()
        print("waiting for tasks to be stopped")
        if not self._reader_task.done():
            self._reader_task.cancel()
            await asyncio.wait([self._reader_stopped.wait()], loop=self._loop)
        print("closing writer")
        try:
            await self.writer.close()
        except Exception as e:
            print("Handler writer close failed: %s" % e)

    def _stop_waiters(self):
        print("Stopping %d puback waiters" % len(self._puback_waiters))
        print("Stopping %d pucomp waiters" % len(self._pubcomp_waiters))
        print("Stopping %d purec waiters" % len(self._pubrec_waiters))
        print("Stopping %d purel waiters" % len(self._pubrel_waiters))
        for waiter in itertools.chain(
                self._puback_waiters.values(),
                self._pubcomp_waiters.values(),
                self._pubrec_waiters.values(),
                self._pubrel_waiters.values()
        ):
            waiter.cancel()

    async def _retry_deliveries(self):
        print("Begin messages delivery retries")
        tasks = []
        for message in itertools.chain(self.session.inflight_in.values(), self.session.inflight_out.values()):
            tasks.append(asyncio.wait_for(self._handle_message_flow(message), 10, loop=self._loop))
        if tasks:
            done, pending = await asyncio.wait(tasks, loop=self._loop)
            print("%d messages redelivered" % len(done))
            print("%d messages not redelivered due to timeout" % len(pending))
        print("End messages delivery retries")

    async def mqtt_publish(self, topic, data, qos, retain, ack_timeout=None):
        if qos in (QOS_1, QOS_2):
            packet_id = self.session.next_packet_id
            if packet_id in self.session.inflight_out:
                raise HZMQTTException("A message with the same packet ID '%d' is already in flight" % packet_id)
        else:
            packet_id = None

        message = OutgoingApplicationMessage(packet_id, topic, qos, data, retain)
        if ack_timeout is not None and ack_timeout > 0:
            await asyncio.wait_for(self._handle_message_flow(message), ack_timeout, loop=self._loop)
        else:
            await self._handle_message_flow(message)

        return message

    async def _handle_message_flow(self, app_message):
        if app_message.qos == QOS_0:
            await self._handle_qos0_message_flow(app_message)
        elif app_message.qos == QOS_1:
            await self._handle_qos1_message_flow(app_message)
        elif app_message.qos == QOS_2:
            await self._handle_qos2_message_flow(app_message)
        else:
            raise HZMQTTException("Unexcepted QOS value '%d" % str(app_message.qos))

    async def _handle_qos0_message_flow(self, app_message):
        assert app_message.qos == QOS_0
        if app_message.direction == OUTGOING:
            packet = app_message.build_publish_packet()
            await self._send_packet(packet)
            app_message.publish_packet = packet
        elif app_message.direction == INCOMING:
            if app_message.publish_packet.dup_flag:
                print("[MQTT-3.3.1-2] DUP flag must set to 0 for QOS 0 message. Message ignored: %s" % repr(app_message.publish_packet))
            else:
                try:
                    self.session.delivered_message_queue.put_nowait(app_message)
                except:
                    print("delivered messages queue full. QOS_0 message discarded")

    async def _handle_qos1_message_flow(self, app_message):
        assert app_message.qos == QOS_1
        if app_message.puback_packet:
            raise HZMQTTException("Message '%d' has already been acknowledged" % app_message.packet_id)
        if app_message.direction == OUTGOING:
            if app_message.packet_id not in self.session.inflight_out:
                self.session.inflight_out[app_message.packet_id] = app_message
            if app_message.publish_packet is not None:
                publish_packet = app_message.build_publish_packet(dup=True)
            else:
                publish_packet = app_message.build_publish_packet()

            await self._send_packet(publish_packet)
            app_message.publish_packet = publish_packet

            waiter = asyncio.Future(loop=self._loop)
            self._puback_waiters[app_message.packet_id] = waiter
            await waiter
            del self._puback_waiters[app_message.packet_id]
            app_message.puback_packet = waiter.result()

            del self.session.inflight_out[app_message.packet_id]
        elif app_message.direction == INCOMING:
            print("Add message to delivery")
            await self.session.delivered_message_queue.put(app_message)

            puback = PubackPacket.build(app_message.packet_id)
            await self._send_packet(puback)
            app_message.puback_packet = puback

    async def _handle_qos2_message_flow(self, app_message):
        assert app_message.qos == QOS_2
        if app_message.direction == OUTGOING:
            if app_message.pubrel_packet and app_message.pubcomp_packet:
                raise HZMQTTException("Message '%d' has already been acknowledged" % app_message.packet_id)
            if not app_message.pubrel_packet:
                if app_message.publish_packet is not None:
                    if app_message.packet_id not in self.session.inflight_out:
                        raise HZMQTTException("Unknown inflight message '%d' in session" % app_message.packet_id)
                    publish_packet = app_message.build_publish_packet(dup=True)
                else:
                    self.session.inflight_out[app_message.packet_id] = app_message
                    publish_packet = app_message.build_publish_packet()

                await self._send_packet(publish_packet)
                app_message.publish_packet = publish_packet

                if app_message.packet_id in self._pubrec_waiters:
                    message = "Can't add PUBREC waiter, a waiter already exists for message Id '%s'" % app_message.packet_id
                    print(message)
                    raise HZMQTTException(message)
                waiter = asyncio.Future(loop=self._loop)
                self._pubrec_waiters[app_message.packet_id] = waiter
                await waiter
                del self._pubrec_waiters[app_message.packet_id]
                app_message.pubrec_packet = waiter.result()
            if not app_message.pubcomp_packet:
                app_message.pubrel_packet = PubrelPacket.build(app_message.packet_id)
                await self._send_packet(app_message.pubrel_packet)

                waiter = asyncio.Future(loop=self._loop)
                self._pubcomp_waiters[app_message.packet_id] = waiter
                await waiter
                del self._pubcomp_waiters[app_message.packet_id]
                app_message.pubcomp_packet = waiter.result()

            del self.session.inflight_out[app_message.packet_id]
        elif app_message.direction == INCOMING:
            self.session.inflight_in[app_message.packet_id] = app_message

            pubrec_packet = PubrecPacket.build(app_message.packet_id)
            await self._send_packet(pubrec_packet)
            app_message.pubrec_packet = pubrec_packet

            if app_message.packet_id in self._pubrel_waiters and not self._pubrel_waiters[app_message.packet_id].done():
                message = "A waiter already exists for message Id '%s', canceling it" % app_message.packet_id
                print(message)
                self._pubrel_waiters[app_message.packet_id].cancel()
            try:
                waiter = asyncio.Future(loop=self._loop)
                self._pubrel_waiters[app_message.packet_id] = waiter
                await waiter
                del self._pubrel_waiters[app_message.packet_id]
                app_message.pubrel_packet = waiter.result()

                await self.session.delivered_message_queue.put(app_message)
                del self.session.inflight_in[app_message.packet_id]

                pubcomp_packet = PubcompPacket.build(app_message.packet_id)
                await self._send_packet(pubcomp_packet)
                app_message.pubcomp_packet = pubcomp_packet
            except asyncio.CancelledError:
                print("Message flow cancelled")

    async def _reader_loop(self):
        print("%s Starting reader coro" % self.session.client_id)
        running_tasks = collections.deque()
        keepalive_timeout = self.session.keep_alive
        if keepalive_timeout <= 0:
            keepalive_timeout = None
        while True:
            try:
                self._reader_ready.set()
                while running_tasks and running_tasks[0].done():
                    running_tasks.popleft()
                if len(running_tasks) > 1:
                    print("handler running tasks: %d" % len(running_tasks))

                fixed_header = await asyncio.wait_for(
                    MQTTFixedHeader.from_stream(self.reader),
                    keepalive_timeout, loop=self._loop
                )
                if fixed_header:
                    if fixed_header.packet_type == RESERVED_0 or fixed_header.packet_type == RESERVED_15:
                        print("%s Received reserved packet, which is forbidden: closing connection" % self.session.client_id)
                        await self.handle_connection_closed()
                    else:
                        cls = packet_class(fixed_header)
                        packet = await cls.from_stream(self.reader, fixed_header=fixed_header)
                        print(f'EVENT_MQTT_PACKET_RECEIVED: {packet}')
                        task = None
                        if packet.fixed_header.packet_type == CONNACK:
                            task = asyncio.ensure_future(self.handle_connack(packet), loop=self._loop)
                        elif packet.fixed_header.packet_type == SUBSCRIBE:
                            task = asyncio.ensure_future(self.handle_subscribe(packet), loop=self._loop)
                        elif packet.fixed_header.packet_type == UNSUBSCRIBE:
                            task = asyncio.ensure_future(self.handle_unsubscribe(packet), loop=self._loop)
                        elif packet.fixed_header.packet_type == SUBACK:
                            task = asyncio.ensure_future(self.handle_suback(packet), loop=self._loop)
                        elif packet.fixed_header.packet_type == UNSUBACK:
                            task = asyncio.ensure_future(self.handle_unsuback(packet), loop=self._loop)
                        elif packet.fixed_header.packet_type == PUBACK:
                            task = asyncio.ensure_future(self.handle_puback(packet), loop=self._loop)
                        elif packet.fixed_header.packet_type == PUBREC:
                            task = asyncio.ensure_future(self.handle_pubrec(packet), loop=self._loop)
                        elif packet.fixed_header.packet_type == PUBREL:
                            task = asyncio.ensure_future(self.handle_pubrel(packet), loop=self._loop)
                        elif packet.fixed_header.packet_type == PUBCOMP:
                            task = asyncio.ensure_future(self.handle_pubcomp(packet), loop=self._loop)
                        elif packet.fixed_header.packet_type == PINGREQ:
                            task = asyncio.ensure_future(self.handle_pingreq(packet), loop=self._loop)
                        elif packet.fixed_header.packet_type == PINGRESP:
                            task = asyncio.ensure_future(self.handle_pingresp(packet), loop=self._loop)
                        elif packet.fixed_header.packet_type == PUBLISH:
                            task = asyncio.ensure_future(self.handle_publish(packet), loop=self._loop)
                        elif packet.fixed_header.packet_type == DISCONNECT:
                            task = asyncio.ensure_future(self.handle_disconnect(packet), loop=self._loop)
                        elif packet.fixed_header.packet_type == CONNECT:
                            self.handle_connect(packet)
                        else:
                            print("%s Unhandled packet type: %s" % (self.session.client_id, packet.fixed_header.packet_type))
                        if task:
                            running_tasks.append(task)
                else:
                    print("No more data (EOF received), stopping reader coro")
                    break
            except MQTTException:
                print("Message discarded")
            except asyncio.CancelledError:
                print("Task cancelled, reader loop ending")
                break
            except asyncio.TimeoutError:
                print("Input stream read timeout")
                self.handle_read_timeout()
            except NoDataException:
                print("No data available")
            except BaseException as e:
                print("%s Unhandled exception in reader coro: %r" % (type(self).__name__, e))
                break
        while running_tasks:
            running_tasks.popleft().cancel()
        await self.handle_connection_closed()
        self._reader_stopped.set()
        print("Reader coro stopped")
        await self.stop()

    async def _send_packet(self, packet):
        try:
            with (await self._write_lock):
                await packet.to_stream(self.writer)
            if self._keepalive_task:
                self._keepalive_task.cancel()
                self._keepalive_task = self._loop.call_later(self.keepalive_timeout, self.handle_write_timeout)

            print(f'EVENT_MQTT_PACKET_SENT: {packet}')
        except (ConnectionResetError, BrokenPipeError):
            await self.handle_connection_closed()
        except asyncio.CancelledError:
            raise
        except BaseException as e:
            print("Unhandled exception: %s" % e)
            raise

    async def mqtt_deliver_next_message(self):
        if not self._is_attached():
            return None
        if self.logger.isEnabledFor(logging.DEBUG):
            print("%d message(s) available for delivery" % self.session.delivered_message_queue.qsize())
        try:
            message = await self.session.delivered_message_queue.get()
        except asyncio.CancelledError:
            message = None
        if self.logger.isEnabledFor(logging.DEBUG):
            print("Delivering message %s" % message)
        return message

    def handle_write_timeout(self):
        print('%s write timeout unhandled' % self.session.client_id)

    def handle_read_timeout(self):
        print('%s read timeout unhandled' % self.session.client_id)

    async def handle_connack(self, connack: ConnackPacket):
        print('%s CONNACK unhandled' % self.session.client_id)

    async def handle_connect(self, connect: ConnectPacket):
        print('%s CONNECT unhandled' % self.session.client_id)

    async def handle_subscribe(self, subscribe: SubscribePacket):
        print('%s SUBSCRIBE unhandled' % self.session.client_id)

    async def handle_unsubscribe(self, subscribe: UnsubscribePacket):
        print('%s UNSUBSCRIBE unhandled' % self.session.client_id)

    async def handle_suback(self, suback: SubackPacket):
        print('%s SUBACK unhandled' % self.session.client_id)

    async def handle_unsuback(self, unsuback: UnsubackPacket):
        print('%s UNSUBACK unhandled' % self.session.client_id)

    async def handle_pingresp(self, pingresp: PingRespPacket):
        print('%s PINGRESP unhandled' % self.session.client_id)

    async def handle_pingreq(self, pingreq: PingReqPacket):
        print('%s PINGREQ unhandled' % self.session.client_id)

    async def handle_disconnect(self, disconnect: DisconnectPacket):
        print('%s DISCONNECT unhandled' % self.session.client_id)

    async def handle_connection_closed(self):
        print('%s Connection closed unhandled' % self.session.client_id)

    async def handle_puback(self, puback: PubackPacket):
        packet_id = puback.variable_header.packet_id
        try:
            waiter = self._puback_waiters[packet_id]
            waiter.set_result(puback)
        except KeyError:
            print("Received PUBACK for unknown pending message Id: '%d'" % packet_id)
        except InvalidStateError:
            print("PUBACK waiter with Id '%d' already done" % packet_id)

    async def handle_pubrec(self, pubrec: PubrecPacket):
        packet_id = pubrec.packet_id
        try:
            waiter = self._pubrec_waiters[packet_id]
            waiter.set_result(pubrec)
        except KeyError:
            print("Received PUBREC for unknown pending message with Id: %d" % packet_id)
        except InvalidStateError:
            print("PUBREC waiter with Id '%d' already done" % packet_id)

    async def handle_pubcomp(self, pubcomp: PubcompPacket):
        packet_id = pubcomp.packet_id
        try:
            waiter = self._pubcomp_waiters[packet_id]
            waiter.set_result(pubcomp)
        except KeyError:
            print("Received PUBCOMP for unknown pending message with Id: %d" % packet_id)
        except InvalidStateError:
            print("PUBCOMP waiter with Id '%d' already done" % packet_id)

    async def handle_pubrel(self, pubrel: PubrelPacket):
        packet_id = pubrel.packet_id
        try:
            waiter = self._pubrel_waiters[packet_id]
            waiter.set_result(pubrel)
        except KeyError:
            print("Received PUBREL for unknown pending message with Id: %d" % packet_id)
        except InvalidStateError:
            print("PUBREL waiter with Id '%d' already done" % packet_id)

    async def handle_publish(self, publish_packet: PublishPacket):
        packet_id = publish_packet.variable_header.packet_id
        qos = publish_packet.qos

        incoming_message = IncomingApplicationMessage(packet_id, publish_packet.topic_name, qos, publish_packet.data, publish_packet.retain_flag)
        incoming_message.publish_packet = publish_packet
        await self._handle_message_flow(incoming_message)
        print("Message queue size: %d" % self.session.delivered_message_queue.qsize())


class ClientProtocolHandler(ProtocolHandler):
    def __init__(self, session: Session = None, loop=None):
        super().__init__(session, loop=loop)
        self._ping_task = None
        self._pingresp_queue = asyncio.Queue(loop=self._loop)
        self._subscriptions_waiter = dict()
        self._unsubscriptions_waiter = dict()
        self._disconnect_waiter = None

    async def start(self):
        await super().start()
        if self._disconnect_waiter is None:
            self._disconnect_waiter = futures.Future(loop=self._loop)

    async def stop(self):
        await super().stop()
        if self._ping_task:
            try:
                print("Cancel ping task")
                self._ping_task.cancel()
            except BaseException:
                pass
        if not self._disconnect_waiter.done():
            self._disconnect_waiter.cancel()
        self._disconnect_waiter = None

    def _build_connect_packet(self):
        vh = ConnectVariableHeader()
        payload = ConnectPayload()

        vh.keep_alive = self.session.keep_alive
        vh.clean_session_flag = self.session.clean_session
        vh.will_retain_flag = self.session.will_retain
        payload.client_id = self.session.client_id

        if self.session.username:
            vh.username_flag = True
            payload.username = self.session.username
        else:
            vh.username_flag = False

        if self.session.password:
            vh.password_flag = True
            payload.password = self.session.password
        else:
            vh.password_flag = False
        if self.session.will_flag:
            vh.will_flag = True
            vh.will_qos = self.session.will_qos
            payload.will_message = self.session.will_message
            payload.will_topic = self.session.will_topic
        else:
            vh.will_flag = False

        packet = ConnectPacket(vh=vh, payload=payload)
        return packet

    async def mqtt_connect(self):
        connect_packet = self._build_connect_packet()
        await self._send_packet(connect_packet)
        connack = await ConnackPacket.from_stream(self.reader)
        print(f'EVENT_MQTT_PACKET_RECEIVED: {connack}')
        return connack.return_code

    def handle_write_timeout(self):
        try:
            if not self._ping_task:
                print("Scheduling Ping")
                self._ping_task = asyncio.ensure_future(self.mqtt_ping())
        except BaseException as be:
            print("Exception ignored in ping task: %r" % be)

    def handle_read_timeout(self):
        pass

    async def mqtt_subscribe(self, topics, packet_id):
        subscribe = SubscribePacket.build(topics, packet_id)
        await self._send_packet(subscribe)

        waiter = futures.Future(loop=self._loop)
        self._subscriptions_waiter[subscribe.variable_header.packet_id] = waiter
        return_codes = await waiter

        del self._subscriptions_waiter[subscribe.variable_header.packet_id]
        return return_codes

    async def handle_suback(self, suback: SubackPacket):
        packet_id = suback.variable_header.packet_id
        try:
            waiter = self._subscriptions_waiter.get(packet_id)
            waiter.set_result(suback.payload.return_codes)
        except KeyError as ke:
            print("Received SUBACK for unknown pending subscription with Id: %s" % packet_id)

    async def mqtt_unsubscribe(self, topics, packet_id):
        unsubscribe = UnsubscribePacket.build(topics, packet_id)
        await self._send_packet(unsubscribe)
        waiter = futures.Future(loop=self._loop)
        self._unsubscriptions_waiter[unsubscribe.variable_header.packet_id] = waiter
        await waiter
        del self._unsubscriptions_waiter[unsubscribe.variable_header.packet_id]

    async def handle_unsuback(self, unsuback: UnsubackPacket):
        packet_id = unsuback.variable_header.packet_id
        try:
            waiter = self._unsubscriptions_waiter.get(packet_id)
            waiter.set_result(None)
        except KeyError:
            print("Received UNSUBACK for unknown pending subscription with Id: %s" % packet_id)

    async def mqtt_disconnect(self):
        disconnect_packet = DisconnectPacket()
        await self._send_packet(disconnect_packet)

    async def mqtt_ping(self):
        ping_packet = PingReqPacket()
        await self._send_packet(ping_packet)
        resp = await self._pingresp_queue.get()
        if self._ping_task:
            self._ping_task = None
        return resp

    async def handle_pingresp(self, pingresp: PingRespPacket):
        await self._pingresp_queue.put(pingresp)

    async def handle_connection_closed(self):
        print("Broker closed connection")
        if not self._disconnect_waiter.done():
            self._disconnect_waiter.set_result(None)

    async def wait_disconnect(self):
        await self._disconnect_waiter


class BrokerProtocolHandler(ProtocolHandler):
    def __init__(self, session: Session = None, loop=None):
        super().__init__(session, loop)
        self._disconnect_waiter = None
        self._pending_subscriptions = Queue(loop=self._loop)
        self._pending_unsubscriptions = Queue(loop=self._loop)

    async def start(self):
        await super().start()
        if self._disconnect_waiter is None:
            self._disconnect_waiter = futures.Future(loop=self._loop)

    async def stop(self):
        await super().stop()
        if self._disconnect_waiter is not None and not self._disconnect_waiter.done():
            self._disconnect_waiter.set_result(None)

    async def wait_disconnect(self):
        return await self._disconnect_waiter

    def handle_write_timeout(self):
        pass

    def handle_read_timeout(self):
        if self._disconnect_waiter is not None and not self._disconnect_waiter.done():
            self._disconnect_waiter.set_result(None)

    async def handle_disconnect(self, disconnect):
        print("Client disconnecting")
        if self._disconnect_waiter and not self._disconnect_waiter.done():
            print("Setting waiter result to %r" % disconnect)
            self._disconnect_waiter.set_result(disconnect)

    async def handle_connection_closed(self):
        await self.handle_disconnect(None)

    async def handle_connect(self, connect: ConnectPacket):
        print('%s [MQTT-3.1.0-2] %s : CONNECT message received during messages handling' % (self.session.client_id, format_client_message(self.session)))
        if self._disconnect_waiter is not None and not self._disconnect_waiter.done():
            self._disconnect_waiter.set_result(None)

    async def handle_pingreq(self, pingreq: PingReqPacket):
        await self._send_packet(PingRespPacket.build())

    async def handle_subscribe(self, subscribe: SubscribePacket):
        subscription = {'packet_id': subscribe.variable_header.packet_id, 'topics': subscribe.payload.topics}
        await self._pending_subscriptions.put(subscription)

    async def handle_unsubscribe(self, unsubscribe: UnsubscribePacket):
        unsubscription = {'packet_id': unsubscribe.variable_header.packet_id, 'topics': unsubscribe.payload.topics}
        await self._pending_unsubscriptions.put(unsubscription)

    async def get_next_pending_subscription(self):
        subscription = await self._pending_subscriptions.get()
        return subscription

    async def get_next_pending_unsubscription(self):
        unsubscription = await self._pending_unsubscriptions.get()
        return unsubscription

    async def mqtt_acknowledge_subscription(self, packet_id, return_codes):
        suback = SubackPacket.build(packet_id, return_codes)
        await self._send_packet(suback)

    async def mqtt_acknowledge_unsubscription(self, packet_id):
        unsuback = UnsubackPacket.build(packet_id)
        await self._send_packet(unsuback)

    async def mqtt_connack_authorize(self, authorize: bool):
        if authorize:
            connack = ConnackPacket.build(self.session.parent, CONNECTION_ACCEPTED)
        else:
            connack = ConnackPacket.build(self.session.parent, NOT_AUTHORIZED)
        await self._send_packet(connack)

    @classmethod
    async def init_from_connect(cls, reader: ReaderAdapter, writer: WriterAdapter, loop=None):
        remote_address, remote_port = writer.get_peer_info()
        connect = await ConnectPacket.from_stream(reader)

        if connect.payload.client_id is None:
            raise MQTTException('[[MQTT-3.1.3-3]] : Client identifier must be present')

        if connect.variable_header.will_flag:
            if connect.payload.will_topic is None or connect.payload.will_message is None:
                raise MQTTException('will flag set, but will topic/message not present in payload')

        if connect.variable_header.reserved_flag:
            raise MQTTException('[MQTT-3.1.2-3] CONNECT reserved flag must be set to 0')
        if connect.proto_name != "MQTT":
            raise MQTTException('[MQTT-3.1.2-1] Incorrect protocol name: "%s"' % connect.proto_name)

        connack = None
        error_msg = None
        if connect.proto_level != 4:
            error_msg = 'Invalid protocol from %s: %d' % (format_client_message(address=remote_address, port=remote_port), connect.proto_level)
            connack = ConnackPacket.build(0, UNACCEPTABLE_PROTOCOL_VERSION)
        elif not connect.username_flag and connect.password_flag:
            connack = ConnackPacket.build(0, BAD_USERNAME_PASSWORD)
        elif connect.username_flag and not connect.password_flag:
            connack = ConnackPacket.build(0, BAD_USERNAME_PASSWORD)
        elif connect.username_flag and connect.username is None:
            error_msg = 'Invalid username from %s' % (format_client_message(address=remote_address, port=remote_port))
            connack = ConnackPacket.build(0, BAD_USERNAME_PASSWORD)
        elif connect.password_flag and connect.password is None:
            error_msg = 'Invalid password %s' % (format_client_message(address=remote_address, port=remote_port))
            connack = ConnackPacket.build(0, BAD_USERNAME_PASSWORD)  # [MQTT-3.2.2-4] session_parent=0
        elif connect.clean_session_flag is False and connect.payload.client_id_is_random:
            error_msg = '[MQTT-3.1.3-8] [MQTT-3.1.3-9] %s: No client Id provided (cleansession=0)' % (format_client_message(address=remote_address, port=remote_port))
            connack = ConnackPacket.build(0, IDENTIFIER_REJECTED)
        if connack is not None:
            await connack.to_stream(writer)
            await writer.close()
            raise MQTTException(error_msg)

        incoming_session = Session(loop)
        incoming_session.client_id = connect.client_id
        incoming_session.clean_session = connect.clean_session_flag
        incoming_session.will_flag = connect.will_flag
        incoming_session.will_retain = connect.will_retain_flag
        incoming_session.will_qos = connect.will_qos
        incoming_session.will_topic = connect.will_topic
        incoming_session.will_message = connect.will_message
        incoming_session.username = connect.username
        incoming_session.password = connect.password
        if connect.keep_alive > 0:
            incoming_session.keep_alive = connect.keep_alive
        else:
            incoming_session.keep_alive = 0

        handler = cls(loop=loop)
        return handler, incoming_session



import re

from joatmon.serializable import Serializable


class Relation(Serializable):
    pattern = re.compile(
        r"""(?P<collection1>.*?)
                             \.
                             (?P<field1>.*?)
                             \(\s*(?P<n1>.*?)\)
                             ->
                             (?P<collection2>.*?)
                             \.
                             (?P<field>.*?)
                             \(\s*(?P<n2>.*?)\)""", re.VERBOSE
    )

    def __init__(self, relation=''):
        super(Relation, self).__init__()

        local_collection, local_field, local_relation, foreign_collection, foreign_field, foreign_relation = Relation.pattern.match(relation).groups()

        self.local_collection = local_collection
        self.local_field = local_field
        self.local_relation = local_relation
        self.foreign_collection = foreign_collection
        self.foreign_field = foreign_field
        self.foreign_relation = foreign_relation



import functools
import inspect

import async_exit_stack

from joatmon import context


def transaction(names):
    def _decorator(func):
        @functools.wraps(func)
        async def _wrapper(*args, **kwargs):
            async with async_exit_stack.AsyncExitStack() as stack:
                for name in names or []:
                    await stack.enter_async_context(context.get_value(name))

                return await func(*args, **kwargs)

        _wrapper.__signature__ = inspect.signature(func)
        return _wrapper

    return _decorator



import numpy as np

from joatmon.nn import functional as f
from joatmon.nn.core import (
    Module,
    Parameter,
    Tensor
)

__all__ = ['BatchNorm']


class BatchNorm(Module):
    def __init__(self, features, eps: float = 1e-5, momentum: float = 0.1, affine: bool = True, track_running_stats: bool = True):
        super(BatchNorm, self).__init__()

        self._features = features
        self._eps = eps
        self._momentum = momentum
        self._affine = affine
        self._track_running_stats = track_running_stats
        self._training = True

        if self._affine:
            self.weight = Parameter(Tensor.from_array(np.ones((features,))))
            self.bias = Parameter(Tensor.from_array(np.zeros((features,))))
        else:
            self.weight = None
            self.bias = None

        if self._track_running_stats:
            self.running_mean = Parameter(Tensor.from_array(np.zeros((features,))))
            self.running_var = Parameter(Tensor.from_array(np.ones((features,))))
            self.num_batches_tracked = 0
        else:
            self.running_mean = None
            self.running_var = None
            self.num_batches_tracked = None

        if self._track_running_stats:
            self.running_mean = Parameter(Tensor.from_array(np.zeros((features,))))
            self.running_var = Parameter(Tensor.from_array(np.ones((features,))))
            self.num_batches_tracked = 0

        self.reset_parameters()

    def reset_running_stats(self):
        if self._track_running_stats:
            self.running_mean = Parameter(Tensor.from_array(np.zeros((self._features,))))
            self.running_var = Parameter(Tensor.from_array(np.ones((self._features,))))
            self.num_batches_tracked = 0

    def reset_parameters(self):
        self.reset_running_stats()
        if self._affine:
            self.weight = Parameter(Tensor.from_array(np.ones((self._features,))))
            self.bias = Parameter(Tensor.from_array(np.zeros((self._features,))))

    def forward(self, inp):
        if self._momentum is None:
            exponential_average_factor = 0.0
        else:
            exponential_average_factor = self._momentum

        if self._training and self._track_running_stats:
            if self.num_batches_tracked is not None:
                self.num_batches_tracked = self.num_batches_tracked + 1
                if self._momentum is None:  # use cumulative moving average
                    exponential_average_factor = 1.0 / float(self.num_batches_tracked)
                else:  # use exponential moving average
                    exponential_average_factor = self._momentum

        return f.batch_norm(
            inp=inp,
            weight=self.weight,
            bias=self.bias,
            running_mean=self.running_mean if not self._training or self._track_running_stats else None,
            running_var=self.running_var if not self._training or self._track_running_stats else None,
            momentum=exponential_average_factor,
            eps=self._eps,
            training=self._training or ((self.running_mean is None) and (self.running_var is None))
        )



from __future__ import print_function

import json

import openai

from joatmon.assistant.task import BaseTask


class Task(BaseTask):
    def __init__(self, api, **kwargs):
        super(Task, self).__init__(api, **kwargs)

    @staticmethod
    def help():
        return {
            "name": "translate",
            "description": "a function for user to translate a message",
            "parameters": {
                "type": "object",
                "properties": {
                    "message": {
                        "type": "string",
                        "description": "message to be translated"
                    }
                },
                "required": ["message"]
            }
        }

    def run(self):
        config = json.loads(open('iva/iva.json', 'r').read())['config']['openai']
        openai.api_key = config['key']

        message = self.kwargs.get('message', None) or self.api.input('what do you want the message to be')

        response = openai.ChatCompletion.create(
            model="gpt-3.5-turbo",
            messages=[
                {"role": "system", "content": "You are a helpful assistant."},
                {"role": "user", "content": 'from now on you are going to translate my input and generate only the translated message'},
                {
                    "role": "agent", "content": 'Understood! From now on, I will translate your input and provide the translated message. '
                                                'Please go ahead and let me know what you '
                                                'would like to be translated.'
                },
                {"role": "user", "content": message}
            ]
        )
        self.api.output(response['choices'][0]['message']['content'])

        if not self.stop_event.is_set():
            self.stop_event.set()



import pygame

from joatmon.system.gui.widgets.widget import Element


class Button(Element):
    def __init__(self, id_):
        super(Button, self).__init__(id_)

        self._text = None
        self._font = pygame.font.SysFont(None, 24)

        self._text_color = (0, 0, 0)
        self._background_color = (255, 255, 255)
        self._hover_color = (127, 127, 127)
        self._click_color = (63, 63, 63)

        self._icon = None

        self.is_hovered = False
        self.is_clicked = False

    def icon(self, icon):
        _icon = pygame.image.load(icon)

        self._width = self._height
        # self._rect = None
        self._rect = pygame.Rect(self._x, self._y, self._width, self._height)

        self._icon = pygame.transform.scale(_icon, (self._width - 10, self._height - 10))

        return self

    def text(self, text):
        self._text = text

        return self

    def style(self, style):
        super(Button, self).style(style)

        self._text_color = style.get('text_color', self._text_color)
        self._hover_color = style.get('hover_color', self._hover_color)
        self._click_color = style.get('click_color', self._click_color)

        return self

    def draw(self, screen):
        color = self._click_color if self.is_clicked else self._hover_color if self.is_hovered else self._background_color

        border_color = self._border.get('color', (0, 0, 0))
        thickness = self._border.get('thickness', 1)
        radius = self._border.get('radius', 1)
        shadow = self._border.get('shadow', 3)

        if self._border:
            border_rect = self._rect.copy()
            border_rect.x -= thickness
            border_rect.y -= thickness
            border_rect.width += 2 * thickness
            border_rect.height += 2 * thickness

            if shadow:
                shadow_rect = border_rect.copy()
                shadow_rect.x += shadow
                shadow_rect.y += shadow
                pygame.draw.rect(screen, pygame.Color('gray'), shadow_rect, border_radius=radius)  # shadow color

                pygame.draw.rect(screen, color, border_rect, border_radius=radius)
            pygame.draw.rect(screen, border_color, border_rect, thickness, border_radius=radius)

        if self._icon:
            icon_rect = self._icon.get_rect(center=self._rect.center)
            screen.blit(self._icon, icon_rect)
        else:
            pygame.draw.rect(screen, color, self._rect)

            text_surface = self._font.render(self._text, True, self._text_color)
            text_rect = text_surface.get_rect(center=self._rect.center)
            screen.blit(text_surface, text_rect)

    def handle_event(self, event):
        if event.type == pygame.MOUSEMOTION:
            self.is_hovered = self._rect.collidepoint(event.pos)
            if self.is_hovered:
                _event = self._events.get('hover', None)
                if _event:
                    _event.fire()
        elif event.type == pygame.MOUSEBUTTONDOWN and self.is_hovered:
            self.is_clicked = True
            _event = self._events.get('click', None)
            if _event:
                _event.fire()
        elif event.type == pygame.MOUSEBUTTONUP:
            self.is_clicked = False



import logging

import yaml

logger = logging.getLogger(__name__)


def not_in_dict_or_none(dictionary, key):
    if key not in dictionary or dictionary[key] is None:
        return True
    else:
        return False


def format_client_message(session=None, address=None, port=None):
    if session:
        return "(client id=%s)" % session.client_id
    elif address is not None and port is not None:
        return "(client @=%s:%d)" % (address, port)
    else:
        return "(unknown client)"


def gen_client_id():
    import random
    gen_id = 'hzmqtt/'

    for i in range(7, 23):
        gen_id += chr(random.randint(0, 74) + 48)
    return gen_id


def read_yaml_config(config_file):
    config = None
    try:
        with open(config_file, 'r') as stream:
            config = yaml.full_load(stream) if hasattr(yaml, 'full_load') else yaml.load(stream)
    except yaml.YAMLError as exc:
        logger.error("Invalid config_file %s: %s" % (config_file, exc))
    return config



import math
from functools import (
    partial,
    update_wrapper
)
from typing import (
    List,
    Tuple,
    Union
)

from joatmon.nn.core import Tensor
from joatmon.nn.utility import _calculate_output_dims

BooleanT = Union[bool]
IntegerT = Union[int]
FloatingT = Union[float]
NumberT = Union[IntegerT, FloatingT]

DataT = Union[BooleanT, IntegerT, FloatingT]
ArrayT = Union[Tuple[DataT], List[DataT], Tuple['ArrayT'], List['ArrayT']]
TensorT = Union[ArrayT, Tuple[Tensor], List[Tensor], Tensor, Tuple['TensorT'], List['TensorT']]
TypeT = Union[str]
ShapeT = Union[Tuple[int], List[int]]
IndexT = Union[int, slice, Tuple[Union[int, slice], ...]]

__all__ = [
    'is_tensor', 'concat_backward', 'stack_backward', 'chunk_backward', 'view_backward', 'index_select_backward',
    'squeeze_backward', 'expand_dim_backward', 'transpose_backward', 'absolute_backward', 'around_backward',
    'floor_backward', 'ceil_backward', 'clip_backward', 'negative_backward', 'summation_backward', 'mean_backward',
    'std_backward', 'var_backward', 'add_backward', 'sub_backward', 'mul_backward', 'div_backward', 'power_backward',
    'clone_backward', 'relu_backward', 'sigmoid_backward', 'softmax_backward', 'tanh_backward', 'dense_backward',
    'conv_backward', 'dropout_backward', 'batch_norm_backward', 'max_pool_backward', 'avg_pool_backward', 'lstm_cell_backward',
    'lstm_backward', 'concat', 'stack', 'chunk', 'view', 'index_select', 'zero', 'one', 'fill', 'squeeze', 'expand_dim',
    'transpose', 'absolute', 'around', 'floor', 'ceil', 'clip', 'negative', 'summation', 'mean', 'std', 'var', 'add',
    'sub', 'mul', 'div', 'power', 'clone', 'detach', 'arange', 'linspace', 'normal', 'uniform', 'rand', 'randint',
    'randn', 'eye', 'empty', 'full', 'zeros', 'ones', 'normal_like', 'uniform_like', 'rand_like', 'randint_like',
    'randn_like', 'eye_like', 'empty_like', 'full_like', 'zeros_like', 'ones_like', 'from_array', 'to_array', 'half',
    'single', 'double', 'cpu', 'gpu', 'relu', 'sigmoid', 'softmax', 'tanh', 'dense', 'conv', 'dropout', 'batch_norm',
    'max_pool', 'avg_pool', 'lstm_cell', 'lstm', 'adam', 'rmsprop'
]


# need to implement inplace

# should have c / c++ codes to use them in functional apis


def wrapped_partial(func, *args, **kwargs):
    partial_func = partial(func, *args, **kwargs)
    update_wrapper(partial_func, func)
    return partial_func


def _check_tensor_devices(*tensors: Tensor):
    iterator = iter(tensors)
    try:
        first = next(iterator)
    except StopIteration:
        return True
    return all(first.device == x.device for x in iterator)


def _check_tensors(*tensors: Tensor):
    if not _check_tensor_devices(*tensors):
        raise ValueError('devices are not matching')

    if len(tensors) == 0:
        raise ValueError('there should be at least one tensor')

    if tensors[0].device not in ('cpu', 'gpu'):
        raise ValueError('device has to be either \'cpu\' or \'gpu\'')


def _get_engine(*_: Union[Tensor, str]):
    import numpy as engine
    return engine


def _set_grad(tensor: Tensor, data):
    if not (tensor.requires_grad and not hasattr(tensor, "retains_grad") and not tensor.is_leaf):
        if not tensor.is_leaf:
            return

        if tensor.grad is None:
            tensor.grad = from_array(data, device=tensor.device)
        else:
            tensor.grad._data = tensor.grad.data + data
    tensor.backward(data)


def _create_tensor(*tensors: Tensor, data, func):
    requires_grad = any(map(lambda x: x.requires_grad, tensors))
    grad_fn = None
    if requires_grad:
        grad_fn = func

    tensor = from_array(data=data, requires_grad=requires_grad, device=tensors[0].device)
    tensor._grad_fn = grad_fn
    if any([x.device == 'gpu' for x in tensors]):
        tensor = tensor.gpu()
    return tensor


def is_tensor(obj: object) -> bool:
    return isinstance(obj, Tensor)


def concat_backward(gradient: Tensor, tensors: List[Tensor], axis: int = 0):
    _check_tensors(*tensors)
    engine = _get_engine(*tensors)

    grad_arrays = engine.split(gradient.data, len(tensors), axis=axis)
    for idx, tensor in enumerate(tensors):
        _set_grad(tensor, data=grad_arrays[idx] * engine.ones_like(tensor.data))


def stack_backward(gradient: Tensor, tensors: List[Tensor], axis: int = 0):
    _check_tensors(*tensors)
    engine = _get_engine(*tensors)

    grad_arrays = engine.split(gradient.data, len(tensors), axis=axis)
    for idx, tensor in enumerate(tensors):
        _set_grad(tensor, data=grad_arrays[idx] * engine.ones_like(tensor.data))


def chunk_backward(gradient: Tensor, tensor: Tensor, chunks: int):
    _check_tensors(tensor)
    engine = _get_engine(tensor)

    _set_grad(tensor, gradient.data * engine.ones_like(tensor.data) / chunks)


def view_backward(gradient: Tensor, inp: Tensor):
    _check_tensors(inp)

    _set_grad(inp, gradient.data.reshape(inp.shape))


def index_select_backward(gradient: Tensor, inp: Tensor, index: Tensor, dim: int):
    _check_tensors(inp)
    engine = _get_engine(inp)

    unique, counts = engine.unique(index.data.astype('int'), return_counts=True)
    count_dict = dict(zip(unique, counts))

    index_array = engine.asarray([val for val in range(inp.size(dim))]).astype('int')
    count_array = engine.asarray([count_dict.get(val, 0) for val in range(inp.size(dim))])

    grad_array = engine.zeros_like(gradient.data)
    engine.put_along_axis(grad_array, index_array, count_array, axis=dim)
    _set_grad(inp, data=grad_array)


def squeeze_backward(gradient: Tensor, inp: Tensor):
    _check_tensors(inp)
    engine = _get_engine(inp)

    _set_grad(inp, gradient.data * engine.ones_like(inp.data))


def expand_dim_backward(gradient: Tensor, inp: Tensor):
    _check_tensors(inp)
    engine = _get_engine(inp)

    _set_grad(inp, gradient.data * engine.ones_like(inp.data))


def transpose_backward(gradient: Tensor, inp: Tensor):
    _check_tensors(inp)
    engine = _get_engine(inp)

    _set_grad(inp, gradient.data * engine.ones_like(inp.data))


def absolute_backward(gradient: Tensor, inp: Tensor):
    _check_tensors(inp)
    engine = _get_engine(inp)

    _set_grad(inp, gradient.data * engine.ones_like(inp.data))


def around_backward(gradient: Tensor, inp: Tensor):
    _check_tensors(inp)
    engine = _get_engine(inp)

    _set_grad(inp, gradient.data * engine.ones_like(inp.data))


def floor_backward(gradient: Tensor, inp: Tensor):
    _check_tensors(inp)
    engine = _get_engine(inp)

    _set_grad(inp, gradient.data * engine.ones_like(inp.data))


def ceil_backward(gradient: Tensor, inp: Tensor):
    _check_tensors(inp)
    engine = _get_engine(inp)

    _set_grad(inp, gradient.data * engine.ones_like(inp.data))


def clip_backward(gradient: Tensor, inp: Tensor):
    _check_tensors(inp)
    engine = _get_engine(inp)

    _set_grad(inp, gradient.data * engine.ones_like(inp.data))


def negative_backward(gradient: Tensor, inp: Tensor):
    _check_tensors(inp)
    engine = _get_engine(inp)

    _set_grad(inp, gradient.data * engine.ones_like(inp.data))


def summation_backward(gradient: Tensor, inp: Tensor):
    _check_tensors(inp)
    engine = _get_engine(inp)

    _set_grad(inp, gradient.data * engine.ones_like(inp.data))


def mean_backward(gradient: Tensor, inp: Tensor):
    _check_tensors(inp)
    engine = _get_engine(inp)

    _set_grad(inp, gradient.data * engine.ones_like(inp.data))


def std_backward(gradient: Tensor, inp: Tensor):
    _check_tensors(inp)
    engine = _get_engine(inp)

    _set_grad(inp, gradient.data * engine.ones_like(inp.data))


def var_backward(gradient: Tensor, inp: Tensor):
    _check_tensors(inp)
    engine = _get_engine(inp)

    _set_grad(inp, gradient.data * engine.ones_like(inp.data))


def add_backward(gradient: Tensor, inp1: Tensor, inp2: Tensor):
    _check_tensors(inp1, inp2)
    engine = _get_engine(inp1, inp2)

    _set_grad(inp1, gradient.data * engine.ones_like(inp1.data))
    _set_grad(inp2, gradient.data * engine.ones_like(inp2.data))


def sub_backward(gradient: Tensor, inp1: Tensor, inp2: Tensor):
    _check_tensors(inp1, inp2)
    engine = _get_engine(inp1, inp2)

    _set_grad(inp1, gradient.data * engine.ones_like(inp1.data))
    _set_grad(inp2, gradient.data * -engine.ones_like(inp2.data))


def mul_backward(gradient: Tensor, inp1: Tensor, inp2: Tensor):
    _check_tensors(inp1, inp2)

    _set_grad(inp1, gradient.data * inp2.data)
    _set_grad(inp2, gradient.data * inp1.data)


def div_backward(gradient: Tensor, inp1: Tensor, inp2: Tensor):
    _check_tensors(inp1, inp2)

    _set_grad(inp1, gradient.data * (1 / inp2.data))
    _set_grad(inp2, gradient.data * inp1.data)


def power_backward(gradient: Tensor, inp: Tensor, p: int):
    _check_tensors(inp)

    _set_grad(inp, gradient.data * p * (inp.data ** (p - 1)))


def clone_backward(gradient: Tensor, inp: Tensor):
    _check_tensors(inp)
    engine = _get_engine(inp)

    _set_grad(inp, gradient.data * engine.ones_like(inp.data))


def relu_backward(gradient: Tensor, inp: Tensor, alpha: float):
    _check_tensors(inp)
    engine = _get_engine(inp)

    out = engine.where(inp.data > 0, 1, alpha)

    # out[inp.data <= 0] = 0
    # out[inp.data > 0] = 1

    _set_grad(inp, gradient.data * out)


def sigmoid_backward(gradient: Tensor, inp: Tensor, out):
    _check_tensors(inp)

    _set_grad(inp, gradient.data * out * (1 - out))


def softmax_backward(gradient: Tensor, inp: Tensor):
    _check_tensors(inp)
    engine = _get_engine(inp)

    grad_array = gradient.data

    indices = engine.where(grad_array == grad_array.max())

    arr = -grad_array * grad_array
    arr[indices] = grad_array[indices] * (1 - grad_array[indices])

    _set_grad(inp, arr)


def tanh_backward(gradient: Tensor, inp: Tensor, out: Tensor):
    _check_tensors(inp)
    engine = _get_engine(inp)

    _set_grad(inp, gradient.data * (1 - engine.square(out.data)))


def dense_backward(gradient: Tensor, inp: Tensor, weight: Tensor, bias: Tensor):
    _check_tensors(inp, weight, bias)
    engine = _get_engine(inp, weight, bias)

    _set_grad(inp, engine.dot(gradient.data, weight.data))
    _set_grad(weight, engine.dot(gradient.data.T, inp.data))
    _set_grad(bias, engine.sum(gradient.data, axis=0, keepdims=True))


def conv_backward(gradient: Tensor, inp: Tensor, weight: Tensor, bias: Tensor, stride: int, padding: Union[List[int], Tuple[int]]):
    _check_tensors(inp, weight, bias)
    engine = _get_engine(inp, weight, bias)

    _padded_input_array = engine.pad(inp.data, ((0, 0), (0, 0), (padding[0], padding[0]), (padding[1], padding[1])), mode='constant')
    _weight_array = weight.data
    _grad_array = gradient.data

    _, _, _input_height, input_width = inp.shape
    _, _, _kernel_height, _kernel_width = _weight_array.shape
    _, _, _output_height, _output_width = _grad_array.shape
    _output_array = engine.zeros_like(_padded_input_array)

    _weight_grad = engine.zeros_like(_weight_array)
    _bias_grad = _grad_array.sum(axis=(0, 2, 3))

    for _row in range(_output_height):
        for _column in range(_output_width):
            _output_array[:, :, _row * stride:_row * stride + _kernel_height, _column * stride:_column * stride + _kernel_width] += engine.sum(
                _weight_array[None, :, :, :, :] *
                _grad_array[:, :, None, _row:_row + 1, _column:_column + 1],
                axis=1
            )
            _weight_grad += engine.sum(
                _padded_input_array[:, None, :, _row * stride:_row * stride + _kernel_height, _column * stride:_column * stride + _kernel_width] *
                _grad_array[:, :, None, _row:_row + 1, _column:_column + 1],
                axis=0
            )

    _set_grad(inp, _weight_grad)
    _set_grad(weight, _bias_grad)
    _set_grad(bias, _output_array[:, :, padding[0]:padding[0] + _input_height, padding[1]:padding[1] + input_width])


def dropout_backward(gradient: Tensor, inp: Tensor, mask, keep_prob: float):
    _check_tensors(inp)
    engine = _get_engine(inp)

    def apply_mask(array) -> engine.array:
        array *= mask
        array /= keep_prob
        return array

    _set_grad(inp, apply_mask(gradient))


def batch_norm_backward(gradient: Tensor, inp: Tensor, weight: Tensor, bias: Tensor, training: bool, **kwargs):
    _check_tensors(inp)
    engine = _get_engine(inp)

    if training:
        batch_size = inp.data.shape[0]
        weight_by_grad = weight.data * gradient.data
        dxc = weight_by_grad / kwargs['input_standard_deviation']
        dstd = -engine.sum(
            (weight_by_grad * kwargs['input_mean_difference']) / (kwargs['input_standard_deviation'] * kwargs['input_standard_deviation']),
            axis=0
        )
        dvar = 0.5 * dstd / kwargs['input_standard_deviation']
        dxc += (2.0 / batch_size) * kwargs['input_mean_difference'] * dvar
        dmu = engine.sum(dxc, axis=0)

        _set_grad(inp, dxc - dmu / batch_size)
        _set_grad(weight, engine.sum(kwargs['input_mean_over_input_standard_deviation'] * gradient.data, axis=0))
        _set_grad(bias, gradient.data.sum(axis=0))
    else:
        weight_by_grad = weight.data * gradient.data

        _set_grad(inp, weight_by_grad / kwargs['input_standard_deviation'])
        _set_grad(weight, engine.sum(kwargs['input_mean_over_input_standard_deviation'] * gradient.data, axis=0))
        _set_grad(bias, gradient.data.sum(axis=0))


def max_pool_backward(gradient: Tensor, inp: Tensor, kernel_size: Union[List[int], Tuple[int]], stride: int, padding: Union[List[int], Tuple[int]], cache: dict):
    _check_tensors(inp)
    engine = _get_engine(inp)

    grad_array = gradient.data

    _, _, _output_height, _output_width = grad_array.shape
    _kernel_height, _kernel_width = kernel_size

    _padded_input_array = engine.pad(inp.data, ((0, 0), (0, 0), (padding[0], padding[0]), (padding[1], padding[1])), mode='constant')
    _output_array = engine.zeros_like(_padded_input_array)

    for _row in range(_output_height):
        for _column in range(_output_width):
            increment = grad_array[:, :, _row:_row + 1, _column:_column + 1] * cache[(_row, _column)]
            _output_array[:, :, _row * stride:_row * stride + _kernel_height, _column * stride:_column * stride + _kernel_width] += increment

    _set_grad(inp, _output_array[:, :, padding[0]:padding[0] + _output_height - 1, padding[1]:padding[1] + _output_width - 1])


def avg_pool_backward(gradient: Tensor, inp: Tensor, kernel_size: Union[List[int], Tuple[int]], stride: int, padding: Union[List[int], Tuple[int]]):
    _check_tensors(inp)
    engine = _get_engine(inp)

    grad_array = gradient.data

    _, _, _output_height, _output_width = grad_array.shape
    _kernel_height, _kernel_width = kernel_size

    _padded_input_array = engine.pad(inp.data, ((0, 0), (0, 0), (padding[0], padding[0]), (padding[1], padding[1])), mode='constant')
    _output_array = engine.zeros_like(_padded_input_array)

    for _row in range(_output_height):
        for _column in range(_output_width):
            increment = grad_array[:, :, _row:_row + 1, _column:_column + 1] / _kernel_height / _kernel_width
            _output_array[:, :, _row * stride:_row * stride + _kernel_height, _column * stride:_column * stride + _kernel_width] += increment

    _set_grad(inp, _output_array[:, :, padding[0]:padding[0] + _output_height - 1, padding[1]:padding[1] + _output_width - 1])


def lstm_cell_backward(gradient, inp, all_weights, cache):
    engine = _get_engine()

    gradient_array = gradient.data.copy()
    inp_array = inp.data.copy()

    out = cache['out']
    hx = cache['hx']
    cx = cache['cx']
    igates = cache['i']
    fgates = cache['f']
    cgates = cache['c']
    ogates = cache['o']

    batch_size, layer_num, time_sequence, hidden_size = hx.shape

    w_ih_arrays = [x[0].data for x in all_weights]
    w_hh_arrays = [x[1].data for x in all_weights]

    delta_array = engine.zeros((batch_size, layer_num, time_sequence, hidden_size))
    delta_out_array = engine.zeros((batch_size, layer_num, time_sequence, hidden_size))
    d_out_array = engine.zeros((batch_size, layer_num, time_sequence, hidden_size))
    d_state_array = engine.zeros((batch_size, layer_num, time_sequence, hidden_size))
    d_cgate_array = engine.zeros((batch_size, layer_num, time_sequence, hidden_size))
    d_igate_array = engine.zeros((batch_size, layer_num, time_sequence, hidden_size))
    d_fgate_array = engine.zeros((batch_size, layer_num, time_sequence, hidden_size))
    d_ogate_array = engine.zeros((batch_size, layer_num, time_sequence, hidden_size))

    d_inp_array = engine.zeros_like(inp_array)

    d_wih_array = [engine.zeros_like(x[0].data) for x in all_weights]
    d_whh_array = [engine.zeros_like(x[1].data) for x in all_weights]
    d_bih_array = [engine.zeros_like(x[2].data) for x in all_weights]
    d_bhh_array = [engine.zeros_like(x[3].data) for x in all_weights]

    for time in range(time_sequence - 1, -1, -1):
        a = 0
        for layer in range(layer_num - 1, -1, -1):
            w_ih_array = w_ih_arrays[layer]
            w_hh_array = w_hh_arrays[layer]

            if time == time_sequence - 1:
                if layer == layer_num - 1:
                    cell_grad = gradient_array[:, time, :]
                else:
                    cell_grad = a
            else:
                if layer == layer_num - 1:
                    cell_grad = gradient_array[:, time, :]
                else:
                    cell_grad = a

            delta_array[:, layer, time, :] = cell_grad
            if time == time_sequence - 1:
                delta_out = engine.zeros_like(delta_out_array[:, layer, time, :])
            else:
                delta_out = delta_out_array[:, layer, time + 1, :]

            d_out_array[:, layer, time, :] = delta_array[:, layer, time, :] + delta_out

            if time == time_sequence - 1:
                d_next_state = engine.zeros_like(d_state_array[:, layer, time, :])
            else:
                d_next_state = d_state_array[:, layer, time + 1, :]

            if time == time_sequence - 1:
                next_forget = engine.zeros_like(fgates[:, layer, time, :])
            else:
                next_forget = fgates[:, layer, time + 1, :]

            d_state_array[:, layer, time, :] = \
                d_out_array[:, layer, time, :] * ogates[:, layer, time, :] * (1 - engine.tanh(cx[:, layer, time, :]) ** 2) + d_next_state * next_forget

            if time == 0:
                prev_state = engine.zeros_like(cx[:, layer, time, :])
            else:
                prev_state = cx[:, layer, time - 1, :]

            d_cgate = d_state_array[:, layer, time, :] * igates[:, layer, time, :] * (1 - cgates[:, layer, time, :] ** 2)
            d_igate = d_state_array[:, layer, time, :] * cgates[:, layer, time, :] * igates[:, layer, time, :] * (1 - igates[:, layer, time, :])
            d_fgate = d_state_array[:, layer, time, :] * prev_state * fgates[:, layer, time, :] * (1 - fgates[:, layer, time, :])
            d_ogate = d_out_array[:, layer, time, :] * engine.tanh(cx[:, layer, time, :]) * ogates[:, layer, time, :] * (1 - ogates[:, layer, time, :])

            d_cgate_array[:, layer, time, :] = d_cgate
            d_igate_array[:, layer, time, :] = d_igate
            d_fgate_array[:, layer, time, :] = d_fgate
            d_ogate_array[:, layer, time, :] = d_ogate

            d_gates = engine.hstack([d_igate, d_fgate, d_cgate, d_ogate])
            # d_x_array[:, layer, time, :] = (w_ih_array.T @ d_gates.T).T
            a = (w_ih_array.T @ d_gates.T).T

            delta_out_array[:, layer, time, :] = (w_hh_array.T @ d_gates.T).T

            # d_wih = engine.outer(d_gates, inp[:, time, :])
            if layer == 0:
                i = inp_array[:, time, :]
            else:
                i = out[:, layer - 1, time, :]
            # d_wih = d_gates.T @ inp_array[:, time, :]
            d_wih = d_gates.T @ i

            if time == time_sequence - 1:
                d_gates_next = engine.hstack(
                    [
                        engine.zeros_like(d_igate_array[:, layer, time, :]),
                        engine.zeros_like(d_fgate_array[:, layer, time, :]),
                        engine.zeros_like(d_cgate_array[:, layer, time, :]),
                        engine.zeros_like(d_ogate_array[:, layer, time, :])
                    ]
                )
            else:
                d_gates_next = engine.hstack(
                    [
                        d_igate_array[:, layer, time + 1, :],
                        d_fgate_array[:, layer, time + 1, :],
                        d_cgate_array[:, layer, time + 1, :],
                        d_ogate_array[:, layer, time + 1, :]
                    ]
                )
            d_whh = d_gates_next.T @ out[:, layer, time, :]

            d_wih_array[layer] += d_wih
            d_whh_array[layer] += d_whh
            d_bih_array[layer] += engine.sum(d_gates.T, axis=1)
            d_bhh_array[layer] += engine.sum(d_gates.T, axis=1)

        d_inp_array[:, time, :] = a

    _set_grad(inp, d_inp_array)
    for layer in range(layer_num):
        _set_grad(all_weights[layer][0], d_wih_array[layer])
        _set_grad(all_weights[layer][1], d_whh_array[layer])
        _set_grad(all_weights[layer][2], d_bih_array[layer])
        _set_grad(all_weights[layer][3], d_bhh_array[layer])


def lstm_backward(gradient, inp, all_weights, cache):
    engine = _get_engine()

    gradient_array = gradient.data.copy()
    inp_array = inp.data.copy()

    out = cache['out']
    hx = cache['hx']
    cx = cache['cx']
    igates = cache['i']
    fgates = cache['f']
    cgates = cache['c']
    ogates = cache['o']

    batch_size, layer_num, time_sequence, hidden_size = hx.shape

    w_ih_arrays = [x[0].data for x in all_weights]
    w_hh_arrays = [x[1].data for x in all_weights]

    delta_array = engine.zeros((batch_size, layer_num, time_sequence, hidden_size))
    delta_out_array = engine.zeros((batch_size, layer_num, time_sequence, hidden_size))
    d_out_array = engine.zeros((batch_size, layer_num, time_sequence, hidden_size))
    d_state_array = engine.zeros((batch_size, layer_num, time_sequence, hidden_size))
    d_cgate_array = engine.zeros((batch_size, layer_num, time_sequence, hidden_size))
    d_igate_array = engine.zeros((batch_size, layer_num, time_sequence, hidden_size))
    d_fgate_array = engine.zeros((batch_size, layer_num, time_sequence, hidden_size))
    d_ogate_array = engine.zeros((batch_size, layer_num, time_sequence, hidden_size))

    d_inp_array = engine.zeros_like(inp_array)

    d_wih_array = [engine.zeros_like(x[0].data) for x in all_weights]
    d_whh_array = [engine.zeros_like(x[1].data) for x in all_weights]
    d_bih_array = [engine.zeros_like(x[2].data) for x in all_weights]
    d_bhh_array = [engine.zeros_like(x[3].data) for x in all_weights]

    for time in range(time_sequence - 1, -1, -1):
        a = 0
        for layer in range(layer_num - 1, -1, -1):
            w_ih_array = w_ih_arrays[layer]
            w_hh_array = w_hh_arrays[layer]

            if time == time_sequence - 1:
                if layer == layer_num - 1:
                    cell_grad = gradient_array[:, time, :]
                else:
                    cell_grad = a
            else:
                if layer == layer_num - 1:
                    cell_grad = gradient_array[:, time, :]
                else:
                    cell_grad = a

            delta_array[:, layer, time, :] = cell_grad
            if time == time_sequence - 1:
                delta_out = engine.zeros_like(delta_out_array[:, layer, time, :])
            else:
                delta_out = delta_out_array[:, layer, time + 1, :]

            d_out_array[:, layer, time, :] = delta_array[:, layer, time, :] + delta_out

            if time == time_sequence - 1:
                d_next_state = engine.zeros_like(d_state_array[:, layer, time, :])
            else:
                d_next_state = d_state_array[:, layer, time + 1, :]

            if time == time_sequence - 1:
                next_forget = engine.zeros_like(fgates[:, layer, time, :])
            else:
                next_forget = fgates[:, layer, time + 1, :]

            d_state_array[:, layer, time, :] = \
                d_out_array[:, layer, time, :] * ogates[:, layer, time, :] * (1 - engine.tanh(cx[:, layer, time, :]) ** 2) + d_next_state * next_forget

            if time == 0:
                prev_state = engine.zeros_like(cx[:, layer, time, :])
            else:
                prev_state = cx[:, layer, time - 1, :]

            d_cgate = d_state_array[:, layer, time, :] * igates[:, layer, time, :] * (1 - cgates[:, layer, time, :] ** 2)
            d_igate = d_state_array[:, layer, time, :] * cgates[:, layer, time, :] * igates[:, layer, time, :] * (1 - igates[:, layer, time, :])
            d_fgate = d_state_array[:, layer, time, :] * prev_state * fgates[:, layer, time, :] * (1 - fgates[:, layer, time, :])
            d_ogate = d_out_array[:, layer, time, :] * engine.tanh(cx[:, layer, time, :]) * ogates[:, layer, time, :] * (1 - ogates[:, layer, time, :])

            d_cgate_array[:, layer, time, :] = d_cgate
            d_igate_array[:, layer, time, :] = d_igate
            d_fgate_array[:, layer, time, :] = d_fgate
            d_ogate_array[:, layer, time, :] = d_ogate

            d_gates = engine.hstack([d_igate, d_fgate, d_cgate, d_ogate])
            # d_x_array[:, layer, time, :] = (w_ih_array.T @ d_gates.T).T
            a = (w_ih_array.T @ d_gates.T).T

            delta_out_array[:, layer, time, :] = (w_hh_array.T @ d_gates.T).T

            # d_wih = engine.outer(d_gates, inp[:, time, :])
            if layer == 0:
                i = inp_array[:, time, :]
            else:
                i = out[:, layer - 1, time, :]
            # d_wih = d_gates.T @ inp_array[:, time, :]
            d_wih = d_gates.T @ i

            if time == time_sequence - 1:
                d_gates_next = engine.hstack(
                    [
                        engine.zeros_like(d_igate_array[:, layer, time, :]),
                        engine.zeros_like(d_fgate_array[:, layer, time, :]),
                        engine.zeros_like(d_cgate_array[:, layer, time, :]),
                        engine.zeros_like(d_ogate_array[:, layer, time, :])
                    ]
                )
            else:
                d_gates_next = engine.hstack(
                    [
                        d_igate_array[:, layer, time + 1, :],
                        d_fgate_array[:, layer, time + 1, :],
                        d_cgate_array[:, layer, time + 1, :],
                        d_ogate_array[:, layer, time + 1, :]
                    ]
                )
            d_whh = d_gates_next.T @ out[:, layer, time, :]

            d_wih_array[layer] += d_wih
            d_whh_array[layer] += d_whh
            d_bih_array[layer] += engine.sum(d_gates.T, axis=1)
            d_bhh_array[layer] += engine.sum(d_gates.T, axis=1)

        d_inp_array[:, time, :] = a

    _set_grad(inp, d_inp_array)
    for layer in range(layer_num):
        _set_grad(all_weights[layer][0], d_wih_array[layer])
        _set_grad(all_weights[layer][1], d_whh_array[layer])
        _set_grad(all_weights[layer][2], d_bih_array[layer])
        _set_grad(all_weights[layer][3], d_bhh_array[layer])


def concat(tensors: List[Tensor], axis: int = 0) -> 'Tensor':
    _check_tensors(*tensors)
    engine = _get_engine(*tensors)

    return _create_tensor(
        *tensors,
        data=engine.concatenate(list(map(lambda x: x.data, tensors)), axis=axis),
        func=wrapped_partial(concat_backward, tensors=tensors, axis=axis)
    )


def stack(tensors: List[Tensor], axis: int = 0) -> 'Tensor':
    _check_tensors(*tensors)
    engine = _get_engine(*tensors)

    return _create_tensor(
        *tensors,
        data=engine.stack(list(map(lambda x: x.data, tensors)), axis=axis),
        func=wrapped_partial(stack_backward, tensors=tensors, axis=axis)
    )


def chunk(tensor: Tensor, chunks: int, dim: int = 0):
    _check_tensors(tensor)
    engine = _get_engine(tensor)

    arrays = engine.split(tensor.data, chunks, dim)

    tensors = []
    for array in arrays:
        tensors.append(
            _create_tensor(
                tensor,
                data=array,
                func=wrapped_partial(chunk_backward, tensor=tensor, chunks=chunks)
            )
        )
    return tensors


def view(inp, size=None) -> 'Tensor':
    _check_tensors(inp)

    return _create_tensor(
        inp,
        data=inp.data.reshape(size),
        func=wrapped_partial(view_backward, inp)
    )


def index_select(inp, dim, index) -> 'Tensor':
    _check_tensors(inp)
    engine = _get_engine(inp)

    return _create_tensor(
        inp,
        data=engine.take_along_axis(inp.data, index.data.astype('int'), dim),
        func=wrapped_partial(index_select_backward, inp=inp, index=index, dim=dim)
    )


def zero(inp) -> 'Tensor':
    _check_tensors(inp)
    engine = _get_engine(inp)

    inp._data = engine.zeros_like(inp.data)
    return inp


def one(inp) -> 'Tensor':
    _check_tensors(inp)
    engine = _get_engine(inp)

    inp._data = engine.ones_like(inp.data)
    return inp


def fill(inp, value) -> 'Tensor':
    _check_tensors(inp)

    inp.data.fill(value)
    return inp


def squeeze(inp, axis=None) -> 'Tensor':
    _check_tensors(inp)
    engine = _get_engine(inp)

    return _create_tensor(
        inp,
        data=engine.squeeze(inp.data, axis=axis),
        func=wrapped_partial(squeeze_backward, inp=inp)
    )


def expand_dim(inp, axis=None) -> 'Tensor':
    _check_tensors(inp)
    engine = _get_engine(inp)

    return _create_tensor(
        inp,
        data=engine.expand_dims(inp.data, axis=axis),
        func=wrapped_partial(expand_dim_backward, inp=inp)
    )


def transpose(inp, axes=None) -> 'Tensor':
    _check_tensors(inp)
    engine = _get_engine(inp)

    return _create_tensor(
        inp,
        data=engine.transpose(inp.data, axes=axes),
        func=wrapped_partial(transpose_backward, inp=inp)
    )


def absolute(inp) -> 'Tensor':
    _check_tensors(inp)
    engine = _get_engine(inp)

    return _create_tensor(
        inp,
        data=engine.absolute(inp.data),
        func=wrapped_partial(absolute_backward, inp=inp)
    )


def around(inp) -> 'Tensor':
    _check_tensors(inp)
    engine = _get_engine(inp)

    return _create_tensor(
        inp,
        data=engine.around(inp.data),
        func=wrapped_partial(around_backward, inp=inp)
    )


def floor(inp) -> 'Tensor':
    _check_tensors(inp)
    engine = _get_engine(inp)

    return _create_tensor(
        inp,
        data=engine.floor(inp.data),
        func=wrapped_partial(floor_backward, inp=inp)
    )


def ceil(inp) -> 'Tensor':
    _check_tensors(inp)
    engine = _get_engine(inp)

    return _create_tensor(
        inp,
        data=engine.ceil(inp.data),
        func=wrapped_partial(ceil_backward, inp=inp)
    )


def clip(inp, min_val, max_val) -> 'Tensor':
    _check_tensors(inp)
    engine = _get_engine(inp)

    return _create_tensor(
        inp,
        data=engine.clip(inp.data, min_val, max_val),
        func=wrapped_partial(clip_backward, inp=inp)
    )


def negative(inp) -> 'Tensor':
    _check_tensors(inp)
    engine = _get_engine(inp)

    return _create_tensor(
        inp,
        data=engine.negative(inp.data),
        func=wrapped_partial(negative_backward, inp=inp)
    )


def summation(inp) -> 'Tensor':
    _check_tensors(inp)
    engine = _get_engine(inp)

    return _create_tensor(
        inp,
        data=engine.sum(inp.data),
        func=wrapped_partial(summation_backward, inp=inp)
    )


def mean(inp) -> 'Tensor':
    _check_tensors(inp)
    engine = _get_engine(inp)

    return _create_tensor(
        inp,
        data=engine.mean(inp.data),
        func=wrapped_partial(mean_backward, inp=inp)
    )


def std(inp) -> 'Tensor':
    _check_tensors(inp)
    engine = _get_engine(inp)

    return _create_tensor(
        inp,
        data=engine.std(inp.data),
        func=wrapped_partial(std_backward, inp=inp)
    )


def var(inp) -> 'Tensor':
    _check_tensors(inp)
    engine = _get_engine(inp)

    return _create_tensor(
        inp,
        data=engine.var(inp.data),
        func=wrapped_partial(var_backward, inp=inp)
    )


def add(inp1, inp2) -> 'Tensor':
    if not isinstance(inp2, Tensor):
        inp2 = from_array(inp2, device=inp1.device)

    _check_tensors(inp1, inp2)

    return _create_tensor(
        inp1,
        inp2,
        data=inp1.data + inp2.data,
        func=wrapped_partial(add_backward, inp1=inp1, inp2=inp2)
    )


def sub(inp1, inp2) -> 'Tensor':
    if not isinstance(inp2, Tensor):
        inp2 = from_array(inp2, device=inp1.device)

    _check_tensors(inp1, inp2)

    return _create_tensor(
        inp1,
        inp2,
        data=inp1.data - inp2.data,
        func=wrapped_partial(sub_backward, inp1=inp1, inp2=inp2)
    )


def mul(inp1, inp2) -> 'Tensor':
    if not isinstance(inp2, Tensor):
        inp2 = from_array(inp2, device=inp1.device)

    _check_tensors(inp1, inp2)

    return _create_tensor(
        inp1,
        inp2,
        data=inp1.data * inp2.data,
        func=wrapped_partial(mul_backward, inp1=inp1, inp2=inp2)
    )


def div(inp1, inp2) -> 'Tensor':
    if not isinstance(inp2, Tensor):
        inp2 = from_array(inp2, device=inp1.device)

    _check_tensors(inp1, inp2)

    return _create_tensor(
        inp1,
        inp2,
        data=inp1.data / inp2.data,
        func=wrapped_partial(div_backward, inp1=inp1, inp2=inp2)
    )


def power(inp, p) -> 'Tensor':
    # if not isinstance(p, Tensor):
    #     p = from_array(p, device=inp.device)

    _check_tensors(inp)

    return _create_tensor(
        inp,
        data=inp.data ** p,
        func=wrapped_partial(power_backward, inp=inp, p=p)
    )


def clone(inp) -> 'Tensor':
    _check_tensors(inp)

    return _create_tensor(
        inp,
        data=inp.data,
        func=wrapped_partial(clone_backward, inp=inp)
    )


def detach(inp, inplace=True) -> 'Tensor':
    _check_tensors(inp)

    if inplace:
        inp._grad_fn = None
        inp._requires_grad = False
        return inp

    _clone = inp.clone()
    _clone._grad_fn = None
    _clone._requires_grad = False
    return _clone


def arange(start=0, stop=0, step=1, requires_grad=False, device='cpu', dtype='float32') -> 'Tensor':
    engine = _get_engine(device)
    return from_array(engine.arange(start, stop, step).astype(dtype), requires_grad, device)


def linspace(start, end, steps, requires_grad=False, device='cpu', dtype='float32') -> 'Tensor':
    engine = _get_engine(device)
    return from_array(engine.linspace(start, end, steps).astype(dtype), requires_grad, device)


def normal(loc=0.0, scale=1.0, size=None, requires_grad=False, device='cpu', dtype='float32') -> 'Tensor':
    engine = _get_engine(device)
    return from_array(engine.random.normal(loc, scale, size).astype(dtype), requires_grad, device)


def uniform(low=-1.0, high=1.0, size=None, requires_grad=False, device='cpu', dtype='float32') -> 'Tensor':
    engine = _get_engine(device)
    return from_array(engine.random.uniform(low, high, size).astype(dtype), requires_grad, device)


def rand(size, requires_grad=False, device='cpu', dtype='float32') -> 'Tensor':
    engine = _get_engine(device)
    return from_array(engine.random.rand(size).astype(dtype), requires_grad, device)


def randint(low=0, high=0, size=None, requires_grad=False, device='cpu', dtype='float32') -> 'Tensor':
    engine = _get_engine(device)
    return from_array(engine.random.randint(low, high, *size).astype(dtype), requires_grad, device)


def randn(size, requires_grad=False, device='cpu', dtype='float32') -> 'Tensor':
    engine = _get_engine(device)
    return from_array(engine.random.randn(size).astype(dtype), requires_grad, device)


def eye(rows, columns=None, requires_grad=False, device='cpu', dtype='float32') -> 'Tensor':
    engine = _get_engine(device)
    return from_array(engine.eye(rows, columns).astype(dtype), requires_grad, device)


def empty(size, requires_grad=False, device='cpu', dtype='float32') -> 'Tensor':
    engine = _get_engine(device)
    return from_array(engine.zeros(size).astype(dtype), requires_grad, device)


def full(size, fill_value, requires_grad=False, device='cpu', dtype='float32') -> 'Tensor':
    engine = _get_engine(device)
    return from_array(engine.full(size, fill_value).astype(dtype), requires_grad, device)


def zeros(size, requires_grad=False, device='cpu', dtype='float32') -> 'Tensor':
    engine = _get_engine(device)
    return from_array(engine.zeros(size).astype(dtype), requires_grad, device)


def ones(size, requires_grad=False, device='cpu', dtype='float32') -> 'Tensor':
    engine = _get_engine(device)
    return from_array(engine.ones(size).astype(dtype), requires_grad, device)


def normal_like(tensor, loc=0.0, scale=1.0, requires_grad=False, device='cpu', dtype='float32') -> 'Tensor':
    return normal(loc, scale, tensor.shape, requires_grad, device, dtype)


def uniform_like(tensor, low=-1.0, high=1.0, requires_grad=False, device='cpu', dtype='float32') -> 'Tensor':
    return uniform(low, high, tensor.shape, requires_grad, device, dtype)


def rand_like(tensor, requires_grad=False, device='cpu', dtype='float32') -> 'Tensor':
    return rand(tensor.shape, requires_grad, device, dtype)


def randint_like(tensor, low=0, high=0, requires_grad=False, device='cpu', dtype='float32') -> 'Tensor':
    return randint(low, high, tensor.shape, requires_grad, device, dtype)


def randn_like(tensor, requires_grad=False, device='cpu', dtype='float32') -> 'Tensor':
    return randn(tensor.shape, requires_grad, device, dtype)


def eye_like(tensor, requires_grad=False, device='cpu', dtype='float32') -> 'Tensor':
    return eye(tensor.shape[0], tensor.shape[1], requires_grad, device, dtype)


def empty_like(tensor, requires_grad=False, device='cpu', dtype='float32') -> 'Tensor':
    return empty(tensor.shape, requires_grad, device, dtype)


def full_like(tensor, fill_value, requires_grad=False, device='cpu', dtype='float32') -> 'Tensor':
    return full(tensor.shape, fill_value, requires_grad, device, dtype)


def zeros_like(tensor, requires_grad=False, device='cpu', dtype='float32') -> 'Tensor':
    return zeros(tensor.shape, requires_grad, device, dtype)


def ones_like(tensor, requires_grad=False, device='cpu', dtype='float32') -> 'Tensor':
    return ones(tensor.shape, requires_grad, device, dtype)


def from_array(data, requires_grad=False, device='cpu', dtype='float32') -> 'Tensor':
    engine = _get_engine('cpu')
    tensor = Tensor(engine.copy(data).astype(dtype), requires_grad=requires_grad)
    if device == 'gpu':
        tensor = tensor.gpu()
    return tensor


def to_array(inp):
    _check_tensors(inp)
    engine = _get_engine('cpu')

    if inp.device != 'cpu':
        raise TypeError('can\'t convert cuda:0 device type tensor to numpy. Use Tensor.cpu() to copy the tensor to host memory first.')
    if inp.requires_grad:
        raise RuntimeError('Can\'t call numpy() on Tensor that requires grad. Use tensor.detach().numpy() instead.')
    return engine.array(inp.data, copy=True)


def half(inp) -> 'Tensor':
    inp._data = inp.data.astype('float16')
    return inp


def single(inp) -> 'Tensor':
    inp._data = inp.data.astype('float32')
    return inp


def double(inp) -> 'Tensor':
    inp._data = inp.data.astype('float64')
    return inp


def cpu(inp) -> 'Tensor':
    if inp.device == 'cpu':
        return inp

    inp._device = 'cpu'
    inp._data = _get_engine().array(inp.data)
    # inp._data = cp.asnumpy(inp.data)
    return inp


def gpu(inp) -> 'Tensor':
    if inp.device == 'gpu':
        return inp

    inp._device = 'gpu'
    inp._data = _get_engine().array(inp.data)
    # inp._data = cp.array(inp.data)
    return inp


def relu(inp, alpha) -> 'Tensor':
    _check_tensors(inp)

    arr = inp.data
    # arr[arr <= 0] = 0
    arr = _get_engine().where(arr > 0, arr, arr * alpha)
    return _create_tensor(
        inp,
        data=arr,
        func=wrapped_partial(relu_backward, inp=inp, alpha=alpha)
    )


def sigmoid(inp) -> 'Tensor':
    _check_tensors(inp)
    engine = _get_engine(inp)

    output_array = 1 / (1 + engine.exp(-inp.data))
    return _create_tensor(
        inp,
        data=output_array,
        func=wrapped_partial(sigmoid_backward, inp=inp, out=output_array)
    )


def softmax(inp) -> 'Tensor':
    _check_tensors(inp)
    engine = _get_engine(inp)

    e = engine.exp(inp.data - inp.data.amax(axis=1, keepdims=True))
    z = e / engine.sum(e, axis=1, keepdims=True)
    return _create_tensor(
        inp,
        data=z,
        func=wrapped_partial(softmax_backward, inp=inp)
    )


def tanh(inp) -> 'Tensor':
    _check_tensors(inp)
    engine = _get_engine(inp)

    output_array = engine.tanh(inp.data)
    return _create_tensor(
        inp,
        data=output_array,
        func=wrapped_partial(tanh_backward, inp=inp, out=output_array)
    )


def dense(inp, weight, bias) -> 'Tensor':
    _check_tensors(inp, weight, bias)
    engine = _get_engine(inp, weight, bias)

    return _create_tensor(
        inp,
        weight,
        bias,
        data=engine.dot(inp.data, weight.data.T) + bias.data,
        func=wrapped_partial(dense_backward, inp=inp, weight=weight, bias=bias)
    )


def conv(inp, weight, bias, stride, padding) -> 'Tensor':
    _check_tensors(inp, weight, bias)
    engine = _get_engine(inp, weight, bias)

    padded_input_array = engine.pad(inp.data, ((0, 0), (0, 0), (padding[0], padding[0]), (padding[1], padding[1])), mode='constant')
    weight_array = weight.data
    bias_array = bias.data

    output_shape = _calculate_output_dims(
        input_shape=inp.shape,
        kernel_shape=weight.shape,
        padding=padding,
        stride=stride
    )

    output_array = engine.zeros(output_shape)

    _, _, kernel_height, kernel_width = weight.shape
    _, _, output_height, output_width = output_shape

    for row in range(output_height):
        for column in range(output_width):
            output_array[:, :, row, column] = engine.sum(
                padded_input_array[:, None, :, row * stride:row * stride + kernel_height, column * stride:column * stride + kernel_width] *
                weight_array[None, :, :, :],
                axis=(2, 3, 4)
            )

    return _create_tensor(
        inp,
        weight,
        bias,
        data=output_array + bias_array[:, None, None],
        func=wrapped_partial(conv_backward, inp=inp, weight=weight, bias=bias, stride=stride, padding=padding)
    )


def dropout(inp, keep_prob) -> 'Tensor':
    _check_tensors(inp)
    engine = _get_engine(inp)

    def apply_mask(array) -> engine.array:
        array *= mask
        array /= keep_prob
        return array

    mask = (engine.random.rand(*inp.shape) < keep_prob)
    out = apply_mask(inp.data)
    return _create_tensor(
        inp,
        data=out,
        func=wrapped_partial(dropout_backward, inp=inp, mask=mask, keep_prob=keep_prob)
    )


def batch_norm(inp, weight, bias, running_mean, running_var, momentum, eps, training) -> 'Tensor':
    _check_tensors(inp)
    engine = _get_engine(inp)

    input_array = inp.data
    running_mean_array = running_mean.data
    running_var_array = running_var.data
    gamma_array = weight.data
    beta_array = bias.data

    if len(input_array.shape) == 2:
        input_mean = engine.mean(input_array, axis=0)

        input_mean_difference = input_array - input_mean
        input_variance = engine.mean(input_mean_difference ** 2, axis=0)
        input_standard_deviation = engine.sqrt(input_variance)
        input_standard_deviation[input_standard_deviation == 0] = input_standard_deviation[input_standard_deviation == 0] + eps
        input_mean_over_input_standard_deviation = input_mean_difference / input_standard_deviation

        if training:
            input_variance[input_variance == 0] = input_variance[input_variance == 0] + eps
            x_hat = (input_array - input_mean) / engine.sqrt(input_variance)
        else:
            running_var_array[running_var_array == 0] = running_var_array[running_var_array == 0] + eps
            x_hat = (input_array - running_mean_array) / engine.sqrt(running_var_array)
        out = gamma_array * x_hat + beta_array
    elif len(input_array.shape) == 4:
        _, channel, _, _ = input_array.shape
        input_mean = engine.mean(input_array, axis=(0, 2, 3))

        input_mean_difference = input_array - input_mean.reshape((1, channel, 1, 1))
        input_variance = engine.mean(input_mean_difference ** 2, axis=(0, 2, 3))
        input_standard_deviation = engine.sqrt(input_variance.reshape((1, channel, 1, 1)))
        input_standard_deviation[input_standard_deviation == 0] = input_standard_deviation[input_standard_deviation == 0] + eps
        input_mean_over_input_standard_deviation = input_mean_difference / input_standard_deviation

        if training:
            input_variance[input_variance == 0] = input_variance[input_variance == 0] + eps
            x_hat = (input_array - input_mean.reshape((1, channel, 1, 1))) / engine.sqrt(input_variance.reshape((1, channel, 1, 1)))
        else:
            running_var_array[running_var_array == 0] = running_var_array[running_var_array == 0] + eps
            x_hat = (input_array - running_mean_array.reshape((1, channel, 1, 1))) / engine.sqrt(running_var_array.reshape((1, channel, 1, 1)))
        out = gamma_array.reshape((1, channel, 1, 1)) * x_hat + beta_array.reshape((1, channel, 1, 1))
    else:
        raise ValueError

    if training:
        running_mean.data = running_mean_array * (1.0 - momentum) + input_mean * momentum
        running_var.data = running_var_array * (1.0 - momentum) + input_variance * momentum

    return _create_tensor(
        inp,
        weight,
        bias,
        data=out,
        func=wrapped_partial(
            batch_norm_backward, inp=inp, weight=weight, bias=bias, training=training, **{
                'input_standard_deviation': input_standard_deviation,
                'input_mean_difference': input_mean_difference,
                'input_mean_over_input_standard_deviation': input_mean_over_input_standard_deviation
            }
        )
    )


def max_pool(inp, kernel_size, stride, padding) -> 'Tensor':
    _check_tensors(inp)
    engine = _get_engine(inp)

    def save_mask(x, cords):
        mask = engine.zeros_like(x)
        n, c, h, w = x.shape
        x = x.reshape(n, h * w, c)
        idx = engine.argmax(x, axis=1)

        n_idx, c_idx = engine.indices((n, c))
        mask.reshape((n, h * w, c))[n_idx, idx, c_idx] = 1
        cache[cords] = mask

    cache = {}

    padded_input_array = engine.pad(inp.data, ((0, 0), (0, 0), (padding[0], padding[0]), (padding[1], padding[1])), mode='constant')

    _, _, output_height, output_width = _calculate_output_dims(inp.shape, (0, 0, kernel_size[0], kernel_size[1]), padding, stride)
    kernel_height, kernel_width = kernel_size
    batch_size, channels, _, _ = padded_input_array.shape

    output_array = engine.zeros((batch_size, channels, output_height, output_width))

    for row in range(output_height):
        for column in range(output_width):
            padded_input_slice = padded_input_array[:, :, row * stride:row * stride + kernel_height, column * stride:column * stride + kernel_width]
            save_mask(x=padded_input_slice, cords=(row, column))
            output_array[:, :, row, column] = engine.amax(padded_input_slice, axis=(2, 3))

    return _create_tensor(
        inp,
        data=output_array,
        func=wrapped_partial(max_pool_backward, inp=inp, kernel_size=kernel_size, stride=stride, padding=padding, cache=cache)
    )


def avg_pool(inp, kernel_size, stride, padding) -> 'Tensor':
    _check_tensors(inp)
    engine = _get_engine(inp)

    padded_input_array = engine.pad(inp.data, ((0, 0), (0, 0), (padding[0], padding[0]), (padding[1], padding[1])), mode='constant')

    _, _, output_height, output_width = _calculate_output_dims(inp.shape, (0, 0, kernel_size[0], kernel_size[1]), padding, stride)
    kernel_height, kernel_width = kernel_size
    batch_size, channels, _, _ = padded_input_array.shape

    output_array = engine.zeros((batch_size, channels, output_height, output_width))

    for row in range(output_height):
        for column in range(output_width):
            padded_input_slice = padded_input_array[:, :, row * stride:row * stride + kernel_height, column * stride:column * stride + kernel_width]
            output_array[:, :, row, column] = engine.mean(padded_input_slice, axis=(2, 3))

    return _create_tensor(
        inp,
        data=output_array,
        func=wrapped_partial(avg_pool_backward, inp=inp, kernel_size=kernel_size, stride=stride, padding=padding)
    )


def lstm_cell(inp, all_weights):
    batch_size, input_features, hidden_size = inp.size(0), inp.size(1), all_weights[0].size(0) // 4

    engine = _get_engine()

    inp_array = inp.data.copy()

    hx_array = engine.zeros((batch_size, hidden_size))
    cx_array = engine.zeros((batch_size, hidden_size))
    out_array = engine.zeros((batch_size, hidden_size))

    igates = engine.zeros((batch_size, hidden_size))
    fgates = engine.zeros((batch_size, hidden_size))
    cgates = engine.zeros((batch_size, hidden_size))
    ogates = engine.zeros((batch_size, hidden_size))

    all_weights_array = [x.data for x in all_weights]

    cell_input = inp_array[:, :]

    w_ih_array, w_hh_array, b_ih_array, b_hh_array = all_weights_array
    h = hx_array[:, :]  # engine.zeros()
    c = cx_array[:, :]  # engine.zeros()

    gates = cell_input @ w_ih_array.T + b_ih_array + h @ w_hh_array.T + b_hh_array

    ingate, forgetgate, cellgate, outgate = engine.split(gates, 4, 1)

    ingate = 1 / (1 + engine.exp(-ingate))
    forgetgate = 1 / (1 + engine.exp(-forgetgate))
    cellgate = engine.tanh(cellgate)
    outgate = 1 / (1 + engine.exp(-outgate))

    igates[:, :] = ingate
    fgates[:, :] = forgetgate
    cgates[:, :] = cellgate
    ogates[:, :] = outgate

    c = (forgetgate * c) + (ingate * cellgate)
    h = outgate * engine.tanh(c)

    hx_array[:, :] = h
    cx_array[:, :] = c

    out_array[:, :] = h

    cache = {
        'inp': inp_array,
        'out': out_array,
        'hx': hx_array,
        'cx': cx_array,
        'i': igates,
        'f': fgates,
        'c': cgates,
        'o': ogates
    }

    out_tensor = _create_tensor(
        *([inp] + all_weights),
        data=out_array[:, :],
        func=wrapped_partial(lstm_cell_backward, inp=inp, all_weights=all_weights, cache=cache)
    )
    return out_tensor, (
        from_array(engine.transpose(hx_array[:, :], (1, 0))),
        from_array(engine.transpose(cx_array[:, :], (1, 0)))
    )


def lstm(inp, all_weights):
    # inp.shape = b, t, f
    # hx.shape = n, b, h
    # cx.shape = n, b, h
    # out.shape = b, t, h

    # ilt = σ(Wxihl−1t + Whihlt−1 + bi) (10)
    # flt = σ(Wxfhl−1t + Whfhlt−1 + bf ) (11)
    # olt = σ(Wxohl−1t + Whohlt−1 + bo) (12)
    # zlt = tanh(Wxzhl−1t + Whzhlt−1 + bz ) (13)
    # clt = flt ◦ clt−1 + ilt ◦ zlt(14)
    # hlt = olt ◦ tanh(clt), (15)

    batch_size, time_sequence, input_features, num_layers, hidden_size = inp.size(0), inp.size(1), inp.size(2), len(all_weights), all_weights[0][0].size(0) // 4

    engine = _get_engine()

    inp_array = inp.data.copy()

    # [[[0] * (4 if layer == 0 else 8) for time in range(time)] for layer in range(layer)]

    hx_array = engine.zeros((batch_size, num_layers, time_sequence, hidden_size))
    cx_array = engine.zeros((batch_size, num_layers, time_sequence, hidden_size))
    out_array = engine.zeros((batch_size, num_layers, time_sequence, hidden_size))

    igates = engine.zeros((batch_size, num_layers, time_sequence, hidden_size))
    fgates = engine.zeros((batch_size, num_layers, time_sequence, hidden_size))
    cgates = engine.zeros((batch_size, num_layers, time_sequence, hidden_size))
    ogates = engine.zeros((batch_size, num_layers, time_sequence, hidden_size))

    w_ih_arrays = [x[0].data for x in all_weights]
    w_hh_arrays = [x[1].data for x in all_weights]
    b_ih_arrays = [x[2].data for x in all_weights]
    b_hh_arrays = [x[3].data for x in all_weights]

    for time in range(time_sequence):
        for layer in range(num_layers):
            if time == 0:
                if layer == 0:
                    cell_input = inp_array[:, time, :]
                else:
                    cell_input = out_array[:, layer - 1, time, :]
            else:
                if layer == 0:
                    cell_input = inp_array[:, time, :]
                else:
                    cell_input = out_array[:, layer - 1, time, :]

            w_ih_array, w_hh_array, b_ih_array, b_hh_array = w_ih_arrays[layer], w_hh_arrays[layer], b_ih_arrays[layer], b_hh_arrays[layer]

            if time == 0:
                h = hx_array[:, layer, 0, :]  # engine.zeros()
                c = cx_array[:, layer, 0, :]  # engine.zeros()
            else:
                h = hx_array[:, layer, time - 1, :]
                c = cx_array[:, layer, time - 1, :]

            gates = cell_input @ w_ih_array.T + b_ih_array + h @ w_hh_array.T + b_hh_array

            ingate, forgetgate, cellgate, outgate = engine.split(gates, 4, 1)

            ingate = 1 / (1 + engine.exp(-ingate))
            forgetgate = 1 / (1 + engine.exp(-forgetgate))
            cellgate = engine.tanh(cellgate)
            outgate = 1 / (1 + engine.exp(-outgate))

            igates[:, layer, time, :] = ingate
            fgates[:, layer, time, :] = forgetgate
            cgates[:, layer, time, :] = cellgate
            ogates[:, layer, time, :] = outgate

            c = (forgetgate * c) + (ingate * cellgate)
            h = outgate * engine.tanh(c)

            hx_array[:, layer, time, :] = h
            cx_array[:, layer, time, :] = c

            out_array[:, layer, time, :] = h

    cache = {
        'inp': inp_array,
        'out': out_array,
        'hx': hx_array,
        'cx': cx_array,
        'i': igates,
        'f': fgates,
        'c': cgates,
        'o': ogates
    }

    out_tensor = _create_tensor(
        *([inp] + sum([layer_weights for layer_weights in all_weights], [])),
        data=out_array[:, num_layers - 1, :, :],
        func=wrapped_partial(lstm_backward, inp=inp, all_weights=all_weights, cache=cache)
    )
    return out_tensor, (
        from_array(engine.transpose(hx_array[:, :, time_sequence - 1, :], (1, 0, 2))),
        from_array(engine.transpose(cx_array[:, :, time_sequence - 1, :], (1, 0, 2)))
    )


def adam(params, grads, exp_avgs, exp_avg_sqs, max_exp_avg_sqs, state_steps, amsgrad, beta1, beta2, lr, weight_decay, eps):
    _check_tensors(*params)
    engine = _get_engine(*params)

    for i, param in enumerate(params):
        grad = grads[i]
        exp_avg = exp_avgs[i]
        exp_avg_sq = exp_avg_sqs[i]
        step = state_steps[i]

        bias_correction1 = 1 - beta1 ** step
        bias_correction2 = 1 - beta2 ** step

        if weight_decay != 0:
            grad._data = grad.data + param.data * weight_decay

        # Decay the first and second moment running average coefficient
        exp_avg._data = exp_avg.data * beta1 + (1 - beta1) * grad.data
        # check if this is true
        exp_avg_sq._data = exp_avg_sq.data * beta2
        exp_avg_sq._data = exp_avg_sq.data + (1 - beta2) * (grad.data * grad.data)
        # exp_avg_sq._data = exp_avg_sq.data * beta2 + (1 - beta2) * (grad.data * grad.data)
        if amsgrad:
            max_exp_avg_sq = max_exp_avg_sqs[i]
            # Maintains the maximum of all 2nd moment running avg. till now
            max_exp_avg_sq._data = engine.maximum(max_exp_avg_sq.data, exp_avg_sq.data)
            # Use the max. for normalizing running avg. of gradient
            denom = (engine.sqrt(max_exp_avg_sq.data) / math.sqrt(bias_correction2)) + eps
        else:
            denom = (engine.sqrt(exp_avg_sq.data) / math.sqrt(bias_correction2)) + eps

        step_size = lr / bias_correction1

        param._data = param.data - step_size * (exp_avg.data / denom)


def rmsprop(params, grads, square_avgs, alphas, momentum_buffers, grad_avgs, momentum, centered, lr, weight_decay, eps):
    _check_tensors(*params)
    engine = _get_engine(*params)

    for i, param in enumerate(params):
        grad = grads[i]
        square_avg = square_avgs[i]
        alpha = alphas[i]

        if weight_decay != 0:
            grad._data = grad.data + param.data * weight_decay

        square_avg._data = square_avg.data * alpha + (1 - alpha) * (grad.data * grad.data)

        if centered:
            grad_avg = grad_avgs[i]
            grad_avg._data = grad_avg.data * alpha + grad.data * (1 - alpha)
            avg = engine.sqrt(square_avg - (grad_avg * grad_avg)) + eps
        else:
            avg = engine.sqrt(square_avg) + eps

        if momentum > 0:
            buf = momentum_buffers[i]
            # check if this is true
            buf._data = buf.data * momentum
            buf._data = buf.data + grad.data / avg
            # buf._data = buf.data * momentum + grad.data / avg
            param._data = param.data - lr * buf.data
        else:
            param._data = param.data - lr * grad.data / avg



import time

__all__ = ['auto_pause']


def auto_pause(duration):
    def _wrapper(func):
        def __wrapper(*args, **kwargs):
            ret = func(*args, **kwargs)
            time.sleep(duration)
            return ret

        return __wrapper

    return _wrapper



import ctypes
import sys

from joatmon.system.decorators import auto_pause

__all__ = ['Keyboard']

if sys.platform != 'win32':
    raise Exception('The keyboard module should only be used on a Windows system.')

SendInput = ctypes.windll.user32.SendInput
MapVirtualKey = ctypes.windll.user32.MapVirtualKeyW

# Constants for failsafe check and pause

FAILSAFE = True
FAILSAFE_POINTS = [(0, 0)]
PAUSE = 0.1  # Tenth-second pause by default.

# Constants for the mouse button names
LEFT = "left"
MIDDLE = "middle"
RIGHT = "right"
PRIMARY = "primary"
SECONDARY = "secondary"

# Mouse Scan Code Mappings
MOUSEEVENTF_MOVE = 0x0001
MOUSEEVENTF_ABSOLUTE = 0x8000
MOUSEEVENTF_LEFTDOWN = 0x0002
MOUSEEVENTF_LEFTUP = 0x0004
MOUSEEVENTF_LEFTCLICK = MOUSEEVENTF_LEFTDOWN + MOUSEEVENTF_LEFTUP
MOUSEEVENTF_RIGHTDOWN = 0x0008
MOUSEEVENTF_RIGHTUP = 0x0010
MOUSEEVENTF_RIGHTCLICK = MOUSEEVENTF_RIGHTDOWN + MOUSEEVENTF_RIGHTUP
MOUSEEVENTF_MIDDLEDOWN = 0x0020
MOUSEEVENTF_MIDDLEUP = 0x0040
MOUSEEVENTF_MIDDLECLICK = MOUSEEVENTF_MIDDLEDOWN + MOUSEEVENTF_MIDDLEUP

# KeyBdInput Flags
KEYEVENTF_EXTENDEDKEY = 0x0001
KEYEVENTF_KEYUP = 0x0002
KEYEVENTF_SCANCODE = 0x0008
KEYEVENTF_UNICODE = 0x0004

# MapVirtualKey Map Types
MAPVK_VK_TO_CHAR = 2
MAPVK_VK_TO_VSC = 0
MAPVK_VSC_TO_VK = 1
MAPVK_VSC_TO_VK_EX = 3

PUL = ctypes.POINTER(ctypes.c_ulong)


class KeyBdInput(ctypes.Structure):
    _fields_ = [("wVk", ctypes.c_ushort),
                ("wScan", ctypes.c_ushort),
                ("dwFlags", ctypes.c_ulong),
                ("time", ctypes.c_ulong),
                ("dwExtraInfo", PUL)]


class HardwareInput(ctypes.Structure):
    _fields_ = [("uMsg", ctypes.c_ulong),
                ("wParamL", ctypes.c_short),
                ("wParamH", ctypes.c_ushort)]


class MouseInput(ctypes.Structure):
    _fields_ = [("dx", ctypes.c_long),
                ("dy", ctypes.c_long),
                ("mouseData", ctypes.c_ulong),
                ("dwFlags", ctypes.c_ulong),
                ("time", ctypes.c_ulong),
                ("dwExtraInfo", PUL)]


class POINT(ctypes.Structure):
    _fields_ = [("x", ctypes.c_long),
                ("y", ctypes.c_long)]


class InputI(ctypes.Union):
    _fields_ = [("ki", KeyBdInput),
                ("mi", MouseInput),
                ("hi", HardwareInput)]


class Input(ctypes.Structure):
    _fields_ = [("type", ctypes.c_ulong),
                ("ii", InputI)]


@auto_pause(duration=0.05)
def _send_keyboard_event(key, event):
    if event == 0x00:
        keybd_flags = KEYEVENTF_SCANCODE

        inserted_events = 0
        expected_events = 1

        if key in [0x4B, 0x48, 0x4D, 0x50]:
            keybd_flags |= KEYEVENTF_EXTENDEDKEY
            if ctypes.windll.user32.GetKeyState(0x90):
                expected_events = 2
                hex_key_code = 0xE0
                extra = ctypes.c_ulong(0)
                ii_ = InputI()
                ii_.ki = KeyBdInput(0, hex_key_code, KEYEVENTF_SCANCODE, 0, ctypes.pointer(extra))
                x = Input(ctypes.c_ulong(1), ii_)

                inserted_events += SendInput(1, ctypes.pointer(x), ctypes.sizeof(x))

        hex_key_code = key
        extra = ctypes.c_ulong(0)
        ii_ = InputI()
        ii_.ki = KeyBdInput(0, hex_key_code, keybd_flags, 0, ctypes.pointer(extra))
        x = Input(ctypes.c_ulong(1), ii_)
        inserted_events += SendInput(1, ctypes.pointer(x), ctypes.sizeof(x))
        return inserted_events == expected_events
    elif event == 0x02:
        keybd_flags = KEYEVENTF_SCANCODE | KEYEVENTF_KEYUP

        inserted_events = 0
        expected_events = 1

        if key in [0x4B, 0x48, 0x4D, 0x50]:
            keybd_flags |= KEYEVENTF_EXTENDEDKEY

        hex_key_code = key
        extra = ctypes.c_ulong(0)
        ii_ = InputI()
        ii_.ki = KeyBdInput(0, hex_key_code, keybd_flags, 0, ctypes.pointer(extra))
        x = Input(ctypes.c_ulong(1), ii_)

        inserted_events += SendInput(1, ctypes.pointer(x), ctypes.sizeof(x))

        if key in [0x4B, 0x48, 0x4D, 0x50] and ctypes.windll.user32.GetKeyState(0x90):
            expected_events = 2
            hex_key_code = 0xE0
            extra = ctypes.c_ulong(0)
            ii_ = InputI()
            ii_.ki = KeyBdInput(0, hex_key_code, KEYEVENTF_SCANCODE | KEYEVENTF_KEYUP, 0, ctypes.pointer(extra))
            x = Input(ctypes.c_ulong(1), ii_)
            inserted_events += SendInput(1, ctypes.pointer(x), ctypes.sizeof(x))

        return inserted_events == expected_events


class Keyboard:
    KEY_DOWN = 0x0000
    KEY_UP = 0x0002

    VK_BACK = (0x08, 0x0000)
    DK_BACK = (0x0E, 0x0008)

    VK_TAB = (0x09, 0x0000)
    DK_TAB = (0x0F, 0x0008)

    VK_CLEAR = (0x0C, 0x0000)
    DK_CLEAR = (0x4C, 0x0008)

    VK_RETURN = (0x0D, 0x0000)
    DK_RETURN = (0x1C, 0x0008)

    VK_SHIFT = (0x10, 0x0000)
    DK_SHIFT = (0x2A, 0x0008)

    VK_CONTROL = (0x11, 0x0000)
    DK_CONTROL = (0x1D, 0x0008)

    VK_MENU = (0x12, 0x0000)
    DK_MENU = (0x38, 0x0008)

    VK_CAPITAL = (0x14, 0x0000)
    DK_CAPITAL = (0x3A, 0x0008)

    VK_ESCAPE = (0x1B, 0x0000)
    DK_ESCAPE = (0x01, 0x0008)

    VK_SPACE = (0x20, 0x0000)
    DK_SPACE = (0x39, 0x0008)

    VK_PRIOR = (0x21, 0x0000)
    DK_PRIOR = (0x49, 0x0008)

    VK_NEXT = (0x22, 0x0000)
    DK_NEXT = (0x51, 0x0008)

    VK_END = (0x23, 0x0000)
    DK_END = (0x4F, 0x0008)

    VK_HOME = (0x24, 0x0000)
    DK_HOME = (0x47, 0x0008)

    VK_LEFT = (0x25, 0x0000)
    DK_LEFT = (0x4B, 0x0008)

    VK_UP = (0x26, 0x0000)
    DK_UP = (0x48, 0x0008)

    VK_RIGHT = (0x27, 0x0000)
    DK_RIGHT = (0x4D, 0x0008)

    VK_DOWN = (0x28, 0x0000)
    DK_DOWN = (0x50, 0x0008)

    VK_SNAPSHOT = (0x2C, 0x0000)
    DK_SNAPSHOT = (0x54, 0x0008)

    VK_INSERT = (0x2D, 0x0000)
    DK_INSERT = (0x52, 0x0008)

    VK_DELETE = (0x2E, 0x0000)
    DK_DELETE = (0x53, 0x0008)

    VK_HELP = (0x2F, 0x0000)
    DK_HELP = (0x63, 0x0008)

    VK_ALPHA_0 = (0x30, 0x0000)
    DK_ALPHA_0 = (0x0B, 0x0008)

    VK_ALPHA_1 = (0x31, 0x0000)
    DK_ALPHA_1 = (0x02, 0x0008)

    VK_ALPHA_2 = (0x32, 0x0000)
    DK_ALPHA_2 = (0x03, 0x0008)

    VK_ALPHA_3 = (0x33, 0x0000)
    DK_ALPHA_3 = (0x04, 0x0008)

    VK_ALPHA_4 = (0x34, 0x0000)
    DK_ALPHA_4 = (0x05, 0x0008)

    VK_ALPHA_5 = (0x35, 0x0000)
    DK_ALPHA_5 = (0x06, 0x0008)

    VK_ALPHA_6 = (0x36, 0x0000)
    DK_ALPHA_6 = (0x07, 0x0008)

    VK_ALPHA_7 = (0x37, 0x0000)
    DK_ALPHA_7 = (0x08, 0x0008)

    VK_ALPHA_8 = (0x38, 0x0000)
    DK_ALPHA_8 = (0x09, 0x0008)

    VK_ALPHA_9 = (0x39, 0x0000)
    DK_ALPHA_9 = (0x0A, 0x0008)

    VK_A = (0x41, 0x0000)
    DK_A = (0x1E, 0x0008)

    VK_B = (0x42, 0x0000)
    DK_B = (0x30, 0x0008)

    VK_C = (0x43, 0x0000)
    DK_C = (0x2E, 0x0008)

    VK_D = (0x44, 0x0000)
    DK_D = (0x20, 0x0008)

    VK_E = (0x45, 0x0000)
    DK_E = (0x12, 0x0008)

    VK_F = (0x46, 0x0000)
    DK_F = (0x21, 0x0008)

    VK_G = (0x47, 0x0000)
    DK_G = (0x22, 0x0008)

    VK_H = (0x48, 0x0000)
    DK_H = (0x23, 0x0008)

    VK_I = (0x49, 0x0000)
    DK_I = (0x17, 0x0008)

    VK_J = (0x4A, 0x0000)
    DK_J = (0x24, 0x0008)

    VK_K = (0x4B, 0x0000)
    DK_K = (0x25, 0x0008)

    VK_L = (0x4C, 0x0000)
    DK_L = (0x26, 0x0008)

    VK_M = (0x4D, 0x0000)
    DK_M = (0x32, 0x0008)

    VK_N = (0x4E, 0x0000)
    DK_N = (0x31, 0x0008)

    VK_O = (0x4F, 0x0000)
    DK_O = (0x18, 0x0008)

    VK_P = (0x50, 0x0000)
    DK_P = (0x19, 0x0008)

    VK_Q = (0x51, 0x0000)
    DK_Q = (0x10, 0x0008)

    VK_R = (0x52, 0x0000)
    DK_R = (0x13, 0x0008)

    VK_S = (0x53, 0x0000)
    DK_S = (0x1F, 0x0008)

    VK_T = (0x54, 0x0000)
    DK_T = (0x14, 0x0008)

    VK_U = (0x55, 0x0000)
    DK_U = (0x16, 0x0008)

    VK_V = (0x56, 0x0000)
    DK_V = (0x2F, 0x0008)

    VK_W = (0x57, 0x0000)
    DK_W = (0x11, 0x0008)

    VK_X = (0x58, 0x0000)
    DK_X = (0x2D, 0x0008)

    VK_Y = (0x59, 0x0000)
    DK_Y = (0x15, 0x0008)

    VK_Z = (0x5A, 0x0000)
    DK_Z = (0x2C, 0x0008)

    VK_NP_0 = (0x60, 0x0000)
    DK_NP_0 = (0x52, 0x0008)

    VK_NP_1 = (0x61, 0x0000)
    DK_NP_1 = (0x4F, 0x0008)

    VK_NP_2 = (0x62, 0x0000)
    DK_NP_2 = (0x50, 0x0008)

    VK_NP_3 = (0x63, 0x0000)
    DK_NP_3 = (0x51, 0x0008)

    VK_NP_4 = (0x64, 0x0000)
    DK_NP_4 = (0x4B, 0x0008)

    VK_NP_5 = (0x65, 0x0000)
    DK_NP_5 = (0x4C, 0x0008)

    VK_NP_6 = (0x66, 0x0000)
    DK_NP_6 = (0x4D, 0x0008)

    VK_NP_7 = (0x67, 0x0000)
    DK_NP_7 = (0x47, 0x0008)

    VK_NP_8 = (0x68, 0x0000)
    DK_NP_8 = (0x48, 0x0008)

    VK_NP_9 = (0x69, 0x0000)
    DK_NP_9 = (0x49, 0x0008)

    VK_MULTIPLY = (0x6A, 0x0000)
    DK_MULTIPLY = (0x37, 0x0008)

    VK_ADD = (0x6B, 0x0000)
    DK_ADD = (0x4E, 0x0008)

    VK_SUBTRACT = (0x6D, 0x0000)
    DK_SUBTRACT = (0x4A, 0x0008)

    VK_DECIMAL = (0x6E, 0x0000)
    DK_DECIMAL = (0x53, 0x0008)

    VK_F1 = (0x70, 0x0000)
    DK_F1 = (0x3B, 0x0008)

    VK_F2 = (0x71, 0x0000)
    DK_F2 = (0x3C, 0x0008)

    VK_F3 = (0x72, 0x0000)
    DK_F3 = (0x3D, 0x0008)

    VK_F4 = (0x73, 0x0000)
    DK_F4 = (0x3E, 0x0008)

    VK_F5 = (0x74, 0x0000)
    DK_F5 = (0x3F, 0x0008)

    VK_F6 = (0x75, 0x0000)
    DK_F6 = (0x40, 0x0008)

    VK_F7 = (0x76, 0x0000)
    DK_F7 = (0x41, 0x0008)

    VK_F8 = (0x77, 0x0000)
    DK_F8 = (0x42, 0x0008)

    VK_F9 = (0x78, 0x0000)
    DK_F9 = (0x43, 0x0008)

    VK_F10 = (0x79, 0x0000)
    DK_F10 = (0x44, 0x0008)

    VK_F11 = (0x7A, 0x0000)
    DK_F11 = (0x57, 0x0008)

    VK_F12 = (0x7B, 0x0000)
    DK_F12 = (0x58, 0x0008)

    VK_F13 = (0x7C, 0x0000)
    DK_F13 = (0x64, 0x0008)

    VK_F14 = (0x7D, 0x0000)
    DK_F14 = (0x65, 0x0008)

    VK_F15 = (0x7E, 0x0000)
    DK_F15 = (0x66, 0x0008)

    VK_F16 = (0x7F, 0x0000)
    DK_F16 = (0x67, 0x0008)

    VK_F17 = (0x80, 0x0000)
    DK_F17 = (0x68, 0x0008)

    VK_F18 = (0x81, 0x0000)
    DK_F18 = (0x69, 0x0008)

    VK_F19 = (0x82, 0x0000)
    DK_F19 = (0x6A, 0x0008)

    VK_F20 = (0x83, 0x0000)
    DK_F20 = (0x6B, 0x0008)

    VK_F21 = (0x84, 0x0000)
    DK_F21 = (0x6C, 0x0008)

    VK_F22 = (0x85, 0x0000)
    DK_F22 = (0x6D, 0x0008)

    VK_F23 = (0x86, 0x0000)
    DK_F23 = (0x6E, 0x0008)

    VK_F24 = (0x87, 0x0000)
    DK_F24 = (0x76, 0x0008)

    VK_NUMLOCK = (0x90, 0x0000)
    DK_NUMLOCK = (0x45, 0x0008)

    VK_SCROLL = (0x91, 0x0000)
    DK_SCROLL = (0x46, 0x0008)

    VK_LSHIFT = (0xA0, 0x0000)
    DK_LSHIFT = (0x2A, 0x0008)

    VK_RSHIFT = (0xA1, 0x0000)
    DK_RSHIFT = (0x36, 0x0008)

    VK_LCONTROL = (0xA2, 0x0000)
    DK_LCONTROL = (0x1D, 0x0008)

    VK_LMENU = (0xA4, 0x0000)
    DK_LMENU = (0x38, 0x0008)

    VK_OEM_1 = (0xBA, 0x0000)
    DK_OEM_1 = (0x27, 0x0008)

    VK_OEM_PLUS = (0xBB, 0x0000)
    DK_OEM_PLUS = (0x0D, 0x0008)

    VK_OEM_COMMA = (0xBC, 0x0000)
    DK_OEM_COMMA = (0x33, 0x0008)

    VK_OEM_MINUS = (0xBD, 0x0000)
    DK_OEM_MINUS = (0x0C, 0x0008)

    VK_OEM_PERIOD = (0xBE, 0x0000)
    DK_OEM_PERIOD = (0x34, 0x0008)

    VK_OEM_2 = (0xBF, 0x0000)
    DK_OEM_2 = (0x35, 0x0008)

    VK_OEM_3 = (0xC0, 0x0000)
    DK_OEM_3 = (0x29, 0x0008)

    VK_OEM_4 = (0xDB, 0x0000)
    DK_OEM_4 = (0x1A, 0x0008)

    VK_OEM_5 = (0xDC, 0x0000)
    DK_OEM_5 = (0x2B, 0x0008)

    VK_OEM_6 = (0xDD, 0x0000)
    DK_OEM_6 = (0x1B, 0x0008)

    VK_OEM_7 = (0xDE, 0x0000)
    DK_OEM_7 = (0x28, 0x0008)

    VK_OEM_102 = (0xE2, 0x0000)
    DK_OEM_102 = (0x56, 0x0008)

    VK_EREOF = (0xF9, 0x0000)
    DK_EREOF = (0x5D, 0x0008)

    VK_ZOOM = (0xFB, 0x0000)
    DK_ZOOM = (0x62, 0x0008)

    @staticmethod
    def key_down(key):
        key, of_type = key
        _send_keyboard_event(key, Keyboard.KEY_DOWN)

    @staticmethod
    def key_up(key):
        key, of_type = key
        _send_keyboard_event(key, Keyboard.KEY_UP)

    @staticmethod
    def press(key):
        Keyboard.key_down(key)
        Keyboard.key_up(key)



import math
from typing import (
    List,
    Tuple
)

import numpy as np

from joatmon.nn import (
    functional as f,
    init
)
from joatmon.nn.core import (
    Module,
    Parameter,
    Tensor
)

__all__ = ['LSTM']


class LSTMCell(Module):
    __constants__ = ['input_size', 'hidden_size']

    input_size: int
    hidden_size: int

    def __init__(self, input_size: int, hidden_size: int) -> None:
        super(LSTMCell, self).__init__()

        self.input_size = input_size
        self.hidden_size = hidden_size

        gate_size = 4 * hidden_size

        self._all_weights = ['weight_ih', 'weight_hh', 'bias_ih', 'bias_hh']
        w_ih = Parameter(Tensor.from_array(np.zeros((gate_size, input_size))))
        w_hh = Parameter(Tensor.from_array(np.zeros((gate_size, hidden_size))))
        b_ih = Parameter(Tensor.from_array(np.zeros((gate_size,))))
        b_hh = Parameter(Tensor.from_array(np.zeros((gate_size,))))
        layer_params = (w_ih, w_hh, b_ih, b_hh)

        for name, param in zip(self._all_weights, layer_params):
            setattr(self, name, param)

        self.reset_parameters()

    def reset_parameters(self) -> None:
        stdv = 1.0 / math.sqrt(self.hidden_size)
        for weight in self.parameters():
            init.uniform(weight, -stdv, stdv)

    def check_input(self, inp: Tensor) -> None:
        expected_input_dim = 3
        if inp.dim() != expected_input_dim:
            raise RuntimeError('input must have {} dimensions, got {}'.format(expected_input_dim, inp.dim()))
        if self.input_size != inp.size(-1):
            raise RuntimeError('input.size(-1) must be equal to input_size. Expected {}, got {}'.format(self.input_size, inp.size(-1)))

    def get_expected_hidden_size(self, inp: Tensor) -> Tuple[int, int]:
        return inp.size(0), self.hidden_size

    def extra_repr(self) -> str:
        s = '{input_size}, {hidden_size}'
        if self.num_layers != 1:
            s += ', num_layers={num_layers}'
        return s.format(**self.__dict__)

    def __setstate__(self, d):
        super(LSTMCell, self).__setstate__(d)
        if 'all_weights' in d:
            self._all_weights = d['all_weights']

        if isinstance(self._all_weights[0], str):
            return

        self._all_weights = ['weight_ih', 'weight_hh', 'bias_ih', 'bias_hh']

    @property
    def all_weights(self) -> List[List[Parameter]]:
        return [getattr(self, weights) for weights in self._all_weights]

    def get_expected_cell_size(self, inp: Tensor) -> Tuple[int, int]:
        return inp.size(0), self.hidden_size

    def check_forward_args(self, inp: Tensor):  # type: ignore
        self.check_input(inp)

    def forward(self, inp):
        self.check_forward_args(inp)
        return f.lstm(inp, self.all_weights)


class LSTM(Module):
    __constants__ = ['input_size', 'hidden_size', 'num_layers']

    input_size: int
    hidden_size: int
    num_layers: int

    def __init__(self, input_size: int, hidden_size: int, num_layers: int = 1) -> None:
        super(LSTM, self).__init__()
        self.input_size = input_size
        self.hidden_size = hidden_size
        self.num_layers = num_layers

        gate_size = 4 * hidden_size

        self._all_weights = []
        for layer in range(num_layers):
            w_ih = Parameter(Tensor.from_array(np.zeros((gate_size, input_size if layer == 0 else hidden_size))))
            w_hh = Parameter(Tensor.from_array(np.zeros((gate_size, hidden_size))))
            b_ih = Parameter(Tensor.from_array(np.zeros((gate_size,))))
            b_hh = Parameter(Tensor.from_array(np.zeros((gate_size,))))
            layer_params = (w_ih, w_hh, b_ih, b_hh)

            param_names = ['weight_ih_l{}', 'weight_hh_l{}']
            param_names += ['bias_ih_l{}', 'bias_hh_l{}']
            param_names = [x.format(layer) for x in param_names]

            for name, param in zip(param_names, layer_params):
                setattr(self, name, param)
            self._all_weights.append(param_names)

        self.reset_parameters()

    def reset_parameters(self) -> None:
        stdv = 1.0 / math.sqrt(self.hidden_size)
        for weight in self.parameters():
            init.uniform(weight, -stdv, stdv)

    def check_input(self, inp: Tensor) -> None:
        expected_input_dim = 3
        if inp.dim() != expected_input_dim:
            raise RuntimeError('input must have {} dimensions, got {}'.format(expected_input_dim, inp.dim()))
        if self.input_size != inp.size(-1):
            raise RuntimeError('input.size(-1) must be equal to input_size. Expected {}, got {}'.format(self.input_size, inp.size(-1)))

    def get_expected_hidden_size(self, inp: Tensor) -> Tuple[int, int, int]:
        return self.num_layers, inp.size(0), self.hidden_size

    def extra_repr(self) -> str:
        s = '{input_size}, {hidden_size}'
        if self.num_layers != 1:
            s += ', num_layers={num_layers}'
        return s.format(**self.__dict__)

    def __setstate__(self, d):
        super(LSTM, self).__setstate__(d)
        if 'all_weights' in d:
            self._all_weights = d['all_weights']

        if isinstance(self._all_weights[0][0], str):
            return
        num_layers = self.num_layers
        self._all_weights = []
        for layer in range(num_layers):
            weights = ['weight_ih_l{}', 'weight_hh_l{}', 'bias_ih_l{}', 'bias_hh_l{}']
            weights = [x.format(layer) for x in weights]
            self._all_weights += [weights[:4]]

    @property
    def all_weights(self) -> List[List[Parameter]]:
        return [[getattr(self, weight) for weight in weights] for weights in self._all_weights]

    def get_expected_cell_size(self, inp: Tensor) -> Tuple[int, int, int]:
        return self.num_layers, inp.size(0), self.hidden_size

    def check_forward_args(self, inp: Tensor):  # type: ignore
        self.check_input(inp)

    def forward(self, inp):
        self.check_forward_args(inp)
        return f.lstm(inp, self.all_weights)



from __future__ import print_function

import time

from joatmon.assistant.service import BaseService
from joatmon.decorator.message import (
    consumer,
    producer
)
from joatmon.plugin.core import register
from joatmon.plugin.message.kafka import KafkaPlugin


class Service(BaseService):
    def __init__(self, api=None, **kwargs):
        super(Service, self).__init__(api, **kwargs)

        register(KafkaPlugin, 'kafka_plugin', 'localhost:9092')

        self.notification = consumer('kafka_plugin', 'notification')(self.notification)
        self.mail = producer('kafka_plugin', 'mail')(self.mail)

    @staticmethod
    def help():
        return {}

    def notification(self, notification):
        subject = notification['subject']
        content_type = notification['content_type']
        content = notification['content']

        if self.kwargs.get('type', None) == 'mail':
            self.mail(
                {
                    "content_type": content_type,
                    "content": content,
                    "subject": subject,
                    "receivers": self.kwargs.get('receivers', [])
                }
            )

    def mail(self, mail):
        ...

    def run(self):
        while True:
            if self.event.is_set():
                break
            time.sleep(1)



from joatmon.nn import functional as f

__all__ = ['AvgPool']


class AvgPool:
    def __init__(self, kernel_size, stride, padding):
        self._kernel_size = kernel_size
        self._stride = stride
        self._padding = padding

    def forward(self, inp):
        return f.avg_pool(
            inp=inp,
            kernel_size=self._kernel_size,
            stride=self._stride,
            padding=self._padding
        )
